#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Tool to substract the temperature effect of given rawdata in a volt.dat file.
The calculation is done after Hayley (2010):

    d_obs^TC = d_obs + (d_est^TC - d_est)

Necessary Input:
* d_obs: measured field data to correct
* d_est: synthetic data of inversion result from d_obs
* d_estTC: synthetic data of temperature corrected inversion result of d_obs

Options:
* output: output file
'''
import numpy as np
from optparse import OptionParser


def handle_options():
    '''Handle options.
    '''
    parser = OptionParser()
    parser.set_defaults(aniso=False)

    parser.add_option("--dobs",
                      dest="d_obs",
                      help="field data",
                      metavar="file",
                      default="mod/volt.dat",
                      )
    parser.add_option("--dest",
                      dest="d_est",
                      help="synthetic data of inversion result",
                      metavar="file",
                      )
    parser.add_option("--desttc",
                      dest="d_estTC",
                      help="synthetic data of corrected inversionr esult",
                      metavar="file",
                      )
    # output options
    parser.add_option("-o",
                      "--output",
                      dest="output",
                      help="output file",
                      metavar="file",
                      default="temp/volt.dat",
                      )

    (options, args) = parser.parse_args()
    return options


def readin_volt(filename):
    """Read in measurement data from a volt.dat file and return electrodes and
    measured resistance.
    """
    with open(filename, 'r') as fid:
        content = np.loadtxt(fid, skiprows=1, usecols=[0, 1, 2])
        volt = content[:, 2]
        elecs = content[:, 0:2]
    return elecs, volt


def calc_correction(volt1, volt2, volt3):
    """Remove the temperature effect from field data using inversion results of
    that data:

        print(volt[0])
        d_obs^TC = d_obs + (d_est^TC - d_est)

    Parameters
    ----------
    d_obs:
        measured field data to correct (volt1)
    d_est:
        synthetic data of inversion result from d_obs (volt2)
    d_estTC:
        synthetic data of temperature corrected inversion result of d_obs
        (volt3)
    """
    volt = np.array([a - b + c for a, b, c in zip(volt1, volt2, volt3)])
    volt[np.where(volt < 0)] = 0.000001

    return volt


def save_volt(elecs, volt, filename):
    """Save the values in volt-format.
    """
    # bring data in shape
    content = np.column_stack((elecs, volt, np.zeros(len(volt))))

    # save datapoints
    with open(filename, 'w') as fid:
        fid.write('{0}\n'.format(content.shape[0]))
    with open(filename, 'ab') as fid:
        np.savetxt(fid, np.array(content), fmt='%i %i %f %f')


def main():
    """Function to remove temperature effect from field data
    """
    options = handle_options()

    # read in observed and synthetic data
    elecs, d_obs = readin_volt(options.d_obs)
    elecs, d_est = readin_volt(options.d_est)
    elecs, d_estTC = readin_volt(options.d_estTC)
    # calculate corrected data
    volt_corr = calc_correction(d_obs,
                                d_est,
                                d_estTC,
                                )
    # save data
    save_volt(elecs,
              volt_corr,
              options.output,
              )


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
import os
import matplotlib.pyplot as plt
import matplotlib as mpl
# from mpl_toolkits.axes_grid1 import AxesGrid


def shiftedColorMap(cmap, start=0, midpoint=0.5, stop=1.0, name='shiftedcmap'):
    '''
    Function to offset the "center" of a colormap. Useful for
    data with a negative min and positive max and you want the
    middle of the colormap's dynamic range to be at zero

    Parameters
    ----------
    cmap:
        The matplotlib colormap to be altered
    start:
        Offset from lowest point in the colormap's range.  Defaults to 0.0 (no
        lower ofset). Should be between 0.0 and `midpoint`.
    midpoint:
        The new center of the colormap. Defaults to 0.5 (no shift). Should be
        between 0.0 and 1.0. In general, this should be  1 - vmax/(vmax +
        abs(vmin)) For example if your data range from -15.0 to +5.0 and you
        want the center of the colormap at 0.0, `midpoint` should be set to  1
        - 5/(5 + 15)) or 0.75
    stop:
        Offset from highets point in the colormap's range.  Defaults to 1.0 (no
        upper ofset). Should be between `midpoint` and 1.0.
    '''
    cdict = {
        'red': [],
        'green': [],
        'blue': [],
        'alpha': []
    }

    # regular index to compute the colors
    reg_index = np.linspace(start, stop, 257)

    # shifted index to match the data
    shift_index = np.hstack([
        np.linspace(0.0, midpoint, 128, endpoint=False),
        np.linspace(midpoint, 1.0, 129, endpoint=True)
    ])

    for ri, si in zip(reg_index, shift_index):
        r, g, b, a = cmap(ri)

        cdict['red'].append((si, r, r))
        cdict['green'].append((si, g, g))
        cdict['blue'].append((si, b, b))
        cdict['alpha'].append((si, a, a))

    newcmap = mpl.colors.LinearSegmentedColormap(name, cdict)
    plt.register_cmap(cmap=newcmap)

    return newcmap


def read_lastmodfile(directory):
    """
    Return the number of the final inversion result.
    """
    filename = '{0}/exe/inv.lastmod'.format(directory)
    # filename HAS to exist. Otherwise the inversion was not finished
    if(not os.path.isfile(filename)):
        return None

    linestring = open(filename, 'r').readline().strip()
    linestring = linestring.replace("\n", '')
    linestring = linestring.replace(".mag", '')
    linestring = linestring.replace("../inv/rho", '')
    return linestring


def open_data():
    content = np.loadtxt('mod/volt.dat', skiprows=1)
    return content  # , array


def open_inv():
    num = read_lastmodfile('.')
    content = np.loadtxt('inv/volt' + num + '.dat', skiprows=1)
    return content  # , array


def main():
    data = open_data()
    inv = open_inv()
    if(
            (data[:, 0].all() - inv[:, 0].all() == 0) and
            (data[:, 1].all() - inv[:, 1].all() == 0)):
        print('Dipoles are equal')
        source = []
        receiver = []
        for AB, MN, rho, pha in data:
            s = ((AB % 1e4) - (round(AB / 1e4))) / 2 + round(AB / 1e4)
            source.append(s)
            r = ((round(MN / 1e4)) - (MN % 1e4)) / 2 + (MN % 1e4)
            receiver.append(r)
        mag_res = np.column_stack(
            (source, receiver, data[:, 2] - inv[:, 2], )
        )

        vmin = min(mag_res[:, 2])
        vmax = max(mag_res[:, 2])
        shrunk_cmap = shiftedColorMap(
            mpl.cm.coolwarm, midpoint=1 - vmax / (vmax + abs(vmin))
        )
        fig, ax = plt.subplots()
        sc = ax.scatter(mag_res[:, 0],
                        mag_res[:, 1],
                        c=mag_res[:, 2],
                        # s=7,
                        edgecolor='k',
                        linewidth=.1,
                        cmap=shrunk_cmap
                        )
        plt.colorbar(sc)
        ax.set_xlabel('source position [m]')
        ax.set_ylabel('receiver position [m]')
        fig.tight_layout()
        fig.savefig('residuals_mag.png', dpi=300)
        pha_res = np.column_stack(
            (source, receiver, data[:, 3] - inv[:, 3], )
        )

        vmin = min(pha_res[:, 2])
        vmax = max(pha_res[:, 2])
        shrunk_cmap = shiftedColorMap(
            mpl.cm.coolwarm, midpoint=1 - vmax / (vmax + abs(vmin))
        )
        fig, ax = plt.subplots()
        sc = ax.scatter(
            pha_res[:, 0],
            pha_res[:, 1],
            c=pha_res[:, 2],
            # s=7,
            edgecolor='k',
            linewidth=.1,
            cmap=shrunk_cmap
        )
        plt.colorbar(sc)
        ax.set_xlabel('source position [m]')
        ax.set_ylabel('receiver position [m]')
        fig.tight_layout()
        fig.savefig('residuals_pha.png', dpi=300)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# *-* coding: utf-8 *-*
"""Look for all unfinished tomodirs in the present directory (and
subdirectories), and run them subsequently by calling CRmod/CRTomo. The number
of cores to use can be specified using a command line option.

Note that this script does not allow any kind of parallel calling of CRTomo, or
remote execution of inversions.

This file is self-contained: it does not need any additional python scripts,
and can thus be also used without installing crtomo-tools. The only requirement
is that CRMod/CRTomo is available in the $PATH variable, i.e., that
CRMod/CRTomo can be called.
"""
import re
import os
import subprocess
import multiprocessing
from optparse import OptionParser


# determine binary paths
try:
    import crtomo.binaries as cBin
    crmod_binary = cBin.get('CRMod')
    crtomo_binary = cBin.get('CRTomo')
except ImportError:
    print('Couldn\'t find crtomo.binaries - falling back to defaults')
    crmod_binary = 'CRMod'
    crtomo_binary = 'CRTomo'


def handle_cmd_options():
    parser = OptionParser()
    parser.add_option(
        '-t', "--threads",
        dest="number_threads",
        type="int",
        help="number of threads EACH CRMod/CRTomo instance uses. If not " +
        "set, will be determined automatically",
        default=None,
    )

    parser.add_option(
        "-n", "--number",
        dest="number_processes",
        help="How many CRMod/CRTomo instances to start in parallel. " +
        "Default: number of detected CPUs/2",
        type='int',
        default=None,
    )

    parser.add_option(
        "-f", "--filter",
        dest="regex_filter",
        help="Process only tomodirs whose path matches the regex",
        type='str',
        default=None,
    )

    parser.add_option(
        "-r", "--reverse",
        dest="reverse_lists",
        help="Reverse directory lists before working with them",
        action='store_true',
    )

    parser.add_option(
        "-c", "--confirm",
        dest="confirm_start",
        help="Prompt for user input before starting",
        action='store_true',
        default=False,
    )

    (options, args) = parser.parse_args()
    return options


def is_tomodir(subdirectories):
    """provided with the subdirectories of a given directory, check if this is
    a tomodir
    """
    required = (
        'exe',
        'config',
        'rho',
        'mod',
        'inv'
    )
    is_tomodir = True
    for subdir in required:
        if subdir not in subdirectories:
            is_tomodir = False
    return is_tomodir


def check_if_needs_modeling(tomodir):
    """check of we need to run CRMod in a given tomodir
    """
    print('check for modeling', tomodir)
    required_files = (
        'config' + os.sep + 'config.dat',
        'rho' + os.sep + 'rho.dat',
        'grid' + os.sep + 'elem.dat',
        'grid' + os.sep + 'elec.dat',
        'exe' + os.sep + 'crmod.cfg',
    )

    not_allowed = (
        'mod' + os.sep + 'volt.dat',
    )
    needs_modeling = True
    for filename in not_allowed:
        if os.path.isfile(tomodir + os.sep + filename):
            needs_modeling = False

    for filename in required_files:
        full_file = tomodir + os.sep + filename
        if not os.path.isfile(full_file):
            print('does not exist: ', full_file)
            needs_modeling = False

    return needs_modeling


def check_if_needs_inversion(tomodir):
    """check of we need to run CRTomo in a given tomodir

    Parameters
    ----------
    tomodir : str
        Tomodir to check

    Returns
    -------
    needs_inversion : bool
        True if not finished yet
    """
    required_files = (
        'grid' + os.sep + 'elem.dat',
        'grid' + os.sep + 'elec.dat',
        'exe' + os.sep + 'crtomo.cfg',
    )

    needs_inversion = True

    for filename in required_files:
        if not os.path.isfile(tomodir + os.sep + filename):
            needs_inversion = False

    # check for crmod OR modeling capabilities
    if not os.path.isfile(tomodir + os.sep + 'mod' + os.sep + 'volt.dat'):
        if not check_if_needs_modeling(tomodir):
            print('no volt.dat and no modeling possible')
            needs_inversion = False

    # check if finished
    inv_ctr_file = tomodir + os.sep + 'inv' + os.sep + 'inv.ctr'
    if os.path.isfile(inv_ctr_file):
        inv_lines = open(inv_ctr_file, 'r').readlines()
        print('inv_lines', inv_lines[-1])
        if inv_lines[-1].startswith('***finished***'):
            needs_inversion = False

    return needs_inversion


def find_unfinished_tomodirs(directory):
    needs_modeling = []
    needs_inversion = []
    for root, dirs, files in os.walk(directory):
        dirs.sort()
        if is_tomodir(dirs):
            print('found tomodir: ', root)
            if check_if_needs_modeling(root):
                needs_modeling.append(root)
            if check_if_needs_inversion(root):
                needs_inversion.append(root)

    return sorted(needs_modeling), sorted(needs_inversion)


def _run_crmod_in_tomodir(tomodir):
    pwd = os.getcwd()
    os.chdir(tomodir + os.sep + 'exe')
    print('Calling CRMod in {}'.format(pwd))
    subprocess.check_output(
        crmod_binary, shell=True, stderr=subprocess.STDOUT, )
    os.chdir(pwd)


def _get_mp_settings(options):
    cpu_count = os.cpu_count()
    if options.number_threads is not None:
        os.environ['OMP_NUM_THREADS'] = '{0}'.format(
            options.number_threads
        )
    else:
        os.environ['OMP_NUM_THREADS'] = '{0}'.format(
            int(cpu_count / 2)
        )

    if options.number_processes is not None:
        number_of_concurrent_processes = options.number_processes
    else:
        number_of_concurrent_processes = int(cpu_count / 2)
    return number_of_concurrent_processes


def run_CRMod(tomodirs, options):
    number_of_concurrent_processes = _get_mp_settings(options)
    pool = multiprocessing.Pool(number_of_concurrent_processes)
    pool.map(_run_crmod_in_tomodir, tomodirs)


def _run_crtomo_in_tomodir(tomodir):
    # check once again if the TD is not yet finished
    if not check_if_needs_inversion(tomodir):
        return
    pwd = os.getcwd()
    os.chdir(tomodir + os.sep + 'exe')
    print('Calling CRTomo in {}'.format(pwd))
    subprocess.check_output(
        crtomo_binary, shell=True, stderr=subprocess.STDOUT, )
    os.chdir(pwd)


def run_CRTomo(tomodirs, options):
    number_of_concurrent_processes = _get_mp_settings(options)
    pool = multiprocessing.Pool(number_of_concurrent_processes)
    pool.map(_run_crtomo_in_tomodir, tomodirs)


def main():
    options = handle_cmd_options()
    needs_modeling, needs_inversion = find_unfinished_tomodirs('.')
    print('-' * 20)
    print('modeling:', needs_modeling)
    print('inversion:', needs_inversion)
    print('-' * 20)

    if options.reverse_lists:
        needs_modeling = reversed(needs_modeling)
        needs_inversion = reversed(needs_inversion)

    if options.regex_filter:
        prog = re.compile(options.regex_filter)
        print('Applying filter:', prog)
        needs_inversion_filtered = list(filter(prog.search, needs_inversion))
        needs_modeling_filtered = list(filter(prog.search, needs_modeling))
    else:
        needs_modeling_filtered = needs_modeling
        needs_inversion_filtered = needs_inversion

    if options.confirm_start:
        print('-' * 20)
        print('Tomodirs to model (after filtering):')
        print(needs_modeling_filtered)
        print('-' * 20)
        print('Tomodirs to invert (after filtering):')
        print(needs_inversion_filtered)
        input('Press any key to proceed')
    run_CRMod(needs_modeling_filtered, options)
    run_CRTomo(needs_inversion_filtered, options)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Create irregular (triangular) grids for CRMod/CRTomo from simple ascii input
files.

Create the input files "electrodes.dat", "boundaries.dat", "char_length.dat"
according to the CRLab manual. An optional file "gmsh_commands.dat" file can be
used to append GMSH commands at the end of the generated gmsh command file.
Note that these command will be executed only after the initial meshing, but
can be used to refine the mesh.

Then run cr_trig_create.py.

If an (optional) parameter is provided, use this as the output directory. This
directory must not exist. In case no parameter is given, create a random
directory using the uuid module.

The output directory is printed in the last line of program output.

Examples:
---------

>>> # note that this is a shell example, not Python!
    cat electrodes.dat
1.0 0.0
2.0 0.0
>>> cat boundaries.dat
0.0 0.0 12
1.0 0.0 12
2.0 0.0 11
2.0 -2.0 11
0.0 -2.0 11
>>> cat char_length.dat
0.5
>>> cat extra_lines.dat
0.5 -0.5 1.5 -1.5
>>> cr_trig_create.py


TODO
----

* this file needs to be restructured. The complexity has outgrown the initial
  structure.
* investigate the ' attractor field' to set characteristic lengths (see notes
  below) http://gmsh.info/doc/texinfo/gmsh.html
"""
import numpy as np
import uuid
import shutil
import os
import subprocess
from optparse import OptionParser

import crtomo.binaries as cBin


def handle_cmd_options():
    parser = OptionParser()
    parser.add_option("-m", "--add_stabilizer_nodes", dest="add_stab_nodes",
                      help="(for surface electrodes only!) How many nodes " +
                      "to put between electrodes", metavar="NR",
                      default=None,
                      type=int)

    (options, args) = parser.parse_args()

    return options, args


class Mesh():
    """
    GMSH distinguishes three types of objects:

    * Points on the boundary (i.e., points that create the boundary)
    * Lines that connect Points to form the actual boundaries
    * Points in Surface, nodes (or points) that lie IN the grid and need to
      be connected to the grid

    Note
    ----

    TODO: I think we don't need to add electrodes as POINTS, only as POINT IN
    SURFACE, because POINTs that do not belong to a line will not be included
    in the final mesh, or? Check.
    Possible Answer: We need to define the Points before we can add them with
    the IN SURFACe command.

    Check for duplicate entries in input files

    Can we test if all electrodes and extra-nodes lie in the boundaries?

    Sort boundaries

    Add extra-nodes (for inner-grid-structure)

    Add extra-lines (for inner-grid-structure)

    """

    def __init__(self):
        self.Points = []
        self.Charlengths = []
        self.Lines = []
        self.Electrodes = []
        self.Boundaries = []
        self.BoundaryIndices = []
        self.ExtraNodes = []
        self.ExtraLineIndices = []

    def get_point_id(self, p, char_length):
        """
        Return the id of the given point (x,y) tuple.
        """
        print('Checking point', p)
        # TODO: This search loop NEEDS to be replaced with something sane
        index = -1
        for nr, i in enumerate(self.Points):
            if(np.all(i == p)):
                print('Point already in list at index {0}'.format(nr))
                if self.Charlengths[nr] > char_length:
                    print('Updating characteristic length')
                    self.Charlengths[nr] = char_length
                return nr

        if(index == -1):
            print('adding point:', p)
            self.Points.append(p)
            self.Charlengths.append(char_length)
            return len(self.Points) - 1

    def add_boundary(self, p1, p2, btype):
        """
        Add a boundary line
        """
        index = self.add_line(p1, p2, self.char_lengths['boundary'])
        # self.Boundaries.append((p1_id,p2_id,btype))
        self.BoundaryIndices.append(index)
        self.Boundaries.append((p1, p2, btype))

    def add_line(self, p1, p2, char_length):
        """
        Add a line to the list. Check if the nodes already exist, and add them
        if not.

        Return the line index (1-indixed, starting with 1)
        """
        p1_id = self.get_point_id(p1, char_length)
        p2_id = self.get_point_id(p2, char_length)
        self.Lines.append((p1_id, p2_id))
        return len(self.Lines)

    def is_in(self, search_list, pair):
        """
        If pair is in search_list, return the index. Otherwise return -1
        """
        index = -1
        for nr, i in enumerate(search_list):
            if(np.all(i == pair)):
                return nr
        return index

    def read_electrodes(self, electrodes):
        """
        Read in electrodes, check if points already exist
        """
        for nr, electrode in enumerate(electrodes):
            index = self.get_point_id(
                electrode, self.char_lengths['electrode'])
            self.Electrodes.append(index)

    def read_extra_nodes(self, filename):
        """Read extra nodes in. Format: x y

        What happens if we add nodes on the boundaries, which are not included
        in the boundaries?
        """
        data = np.atleast_2d(np.loadtxt(filename))
        for nr, pair in enumerate(data):
            index = self.get_point_id(pair, self.char_lengths['extra_node'])
            self.ExtraNodes.append(index)

    def read_extra_lines(self, filename):
        """Read extra lines from the given filename. Each line is defined in
        one line with four coordinates: x1 y1 x2 y2. (x1,y1) denotes the
        starting point, (x2, y2) the end point of the line.
        """
        data = np.atleast_2d(np.loadtxt(filename))
        for nr, coords in enumerate(data):
            p1 = [coords[0], coords[1]]
            p2 = [coords[2], coords[3]]
            index = self.add_line(p1, p2, self.char_lengths['extra_line'])
            self.ExtraLineIndices.append(index)

    def write_electrodes(self, filename):
        """
        Write X Y coordinates of electrodes
        """
        fid = open(filename, 'w')
        for i in self.Electrodes:
            fid.write('{0} {1}\n'.format(self.Points[i][0], self.Points[i][1]))
        fid.close()

    def write_boundaries(self, filename):
        """
        Write boundary lines X1 Y1 X2 Y2 TYPE to file
        """
        fid = open(filename, 'w')
        for i in self.Boundaries:
            print(i)
            # fid.write('{0} {1} {2}\n'.format(i[0], i[1], i[2]))
            fid.write(
                '{0} {1} {2} {3} {4}\n'.format(
                    i[0][0], i[0][1], i[1][0], i[1][1], i[2]))
        fid.close()

    def read_char_lengths(self, filename, electrode_filename):
        """Read characteristic lengths from the given file.

        The file is expected to have either 1 or 4 entries/lines with
        characteristic lengths > 0 (floats). If only one value is encountered,
        it is used for all four entities. If four values are encountered, they
        are assigned, in order, to:

            1) electrode nodes
            2) boundary nodes
            3) nodes from extra lines
            4) nodes from extra nodes

        Note that in case one node belongs to multiple entities, the smallest
        characteristic length will be used.

        If four values are used and the electrode length is negative, then the
        electrode positions will be read in (todo: we open the electrode.dat
        file two times here...) and the minimal distance between all electrodes
        will be multiplied by the absolute value of the imported value, and
        used as the characteristic length:

        .. math::

            l_{electrodes} = min(pdist(electrodes)) * |l_{electrodes}^{from
            file}|

        The function scipy.spatial.distance.pdist is used to compute the global
        minimal distance between any two electrodes.

        It is advisable to only used values in the range [-1, 0) for the
        automatic char length option.
        """

        if os.path.isfile(filename):
            data = np.atleast_1d(np.loadtxt(filename))
            if data.size == 4:
                characteristic_length = data
                # check sign of first (electrode) length value
                if characteristic_length[0] < 0:
                    try:
                        elec_positions = np.loadtxt(electrode_filename)
                    except Exception:
                        raise IOError(
                            'The was an error opening the electrode file')
                    import scipy.spatial.distance
                    distances = scipy.spatial.distance.pdist(elec_positions)
                    characteristic_length[0] = min(distances) * np.abs(
                        characteristic_length[0])
                    if characteristic_length[0] == 0:
                        raise Exception(
                            'Error computing electrode ' +
                            'distances (got a minimal distance of zero')

            else:
                characteristic_length = np.ones(4) * data[0]
        else:
            characteristic_length = np.ones(4)

        if np.any(characteristic_length <= 0):
            raise Exception('No negative characteristic lengths allowed ' +
                            '(except for electrode length')

        self.char_lengths = {}
        for key, item in zip(('electrode',
                              'boundary',
                              'extra_line',
                              'extra_node'),
                             characteristic_length):
            self.char_lengths[key] = item

    def write_points(self, fid):
        """
        Write the grid points to the GMSH-command file.

        Parameters
        ----------
        fid: file object for the command file (.geo)

        """
        for nr, point in enumerate(self.Points):
            fid.write(
                'Point({0}) = {{{1}, {2}, 0, {3}}};\n'.format(
                    nr + 1, point[0], point[1], self.Charlengths[nr]))

    def write_lines(self, fid):
        for nr, line in enumerate(self.Lines):
            fid.write(
                'Line({0}) = {{{1},{2}}};\n'.format(
                    nr + 1, line[0] + 1, line[1] + 1))

    def write_in_plane_nodes(self, fid):
        for nr, line in enumerate(self.Electrodes):
            fid.write('Point {{{0}}} In Surface {{7}};\n'.format(line + 1))

    def write_extra_nodes(self, fid):
        for nr, line in enumerate(self.ExtraNodes):
            fid.write('Point {{{0}}} In Surface {{7}};\n'.format(line + 1))

    def write_geo_file(self, filename):
        """
        Write the .geo file
        """
        fid = open(filename, 'w')
        # 2D mesh algorithm (1=MeshAdapt, 2=Automatic, 5=Delaunay, 6=Frontal,
        # 7=bamg, 8=delquad)
        # according to the GMSH-mailing list the frontal algorithm should be
        # one of the best in terms of grid quality
        fid.write('Mesh.Algorithm = 6;\n')

        self.write_points(fid)
        self.write_lines(fid)

        # fid.write('Coherence;\n')
        # write line loop
        fid.write('Line Loop(1) = {')
        fid.write(','.join(['{0}'.format(x) for x in self.BoundaryIndices]))
        # for i in self.BoundaryIndices:
        #     fid.write('{0},'.format(i))
        fid.write('};\n')
        # # fid.write('{0}}};\n'.format(len(self.Lines)))
        fid.write('Plane Surface(7) = {1};\n')

        self.write_in_plane_nodes(fid)
        # fid.write('Coherence;\n')
        self.write_extra_nodes(fid)
        # fid.write('Coherence;\n')
        for index in self.ExtraLineIndices:
            fid.write('Line {' + '{0}'.format(index) + '} In Surface {7};\n')

        # Lloyd mesh optimisation crashes
        # fid.write('Mesh.Lloyd = 1;\n')

        # run the mesher
        fid.write('Mesh 7;')

        if os.path.isfile('../gmsh_commands.dat'):
            fid2 = open('../gmsh_commands.dat', 'r')
            additional_commands = fid2.read()
            fid2.close()

            fid.write('\n')
            fid.write(additional_commands)

        fid.close()


def check_boundaries(boundaries):
    # generate a complex number for each (x,y) pair
    xy = boundaries[:, 0] + 1j * boundaries[:, 1]
    """
    # for numpy > 1.9 , use:
    unique_values, indices, indices_rev, counts = np.unique(
        xy,
        return_index=True,
        return_inverse=True,
        return_counts=True)
    """
    # numpy 1.8 -->
    unique_values, indices, indices_rev = np.unique(
        xy,
        return_index=True,
        return_inverse=True)
    # now bin the indices we use to reconstruct the original array. Each index
    # that is present more than once will manifest in a bin with a number
    # larger than one. Use as many bins as there are indices.
    nr_bins = np.abs(indices_rev.min() - indices_rev.max()) + 1
    counts, b = np.histogram(indices_rev, bins=nr_bins)
    # <!-- numpy 1.8

    doublets = np.where(counts > 1)
    if doublets[0].size > 0:
        print('ERROR: Duplicate boundary coordinates found!')
        print('ERROR: Debug information')
        for doublet in doublets[0]:
            print('================')
            print('x y type:')
            print(boundaries[doublet, :])
            print('lines: ')
            print(np.where(indices_rev == doublet)[0])
        raise Exception('Duplicate boundary coordinates found!')


def add_stabilizer_nodes(boundaries_raw, electrodes, nr_nodes_between):
    """
    Segmentation of nodes:
        we have the existing nodes
        N.F is the ratio of required nodes and existing nodes
        first, add N nodes to each segment
        then, add one more node to the F first segments

    * assume ordered boundaries
    """
    boundaries = []

    boundaries = boundaries_raw
    # find first electrode in boundary
    for nr in range(electrodes.shape[0] - 1):
        index0 = np.where(
            (boundaries[:, 0] == electrodes[nr, 0]) &
            (boundaries[:, 1] == electrodes[nr, 1])
        )[0]

        index1 = np.where(
            (boundaries[:, 0] == electrodes[nr + 1, 0]) &
            (boundaries[:, 1] == electrodes[nr + 1, 1])
        )[0]

        index0 = index0[0]
        index1 = index1[0]
        if index1 - index0 < 0:
            index0, index1 = index1, index0
        running_index = index0
        nr_nodes = index1 - index0 - 1
        while nr_nodes < nr_nodes_between:
            # determine line equation
            xy0 = boundaries[running_index, 0:2]
            xy1 = boundaries[running_index + 1, 0:2]

            direction = xy1 - xy0
            heading = direction / np.sqrt(np.sum(direction ** 2))

            # new node
            xy_new = xy0 + heading * direction / 2.0
            a = boundaries[running_index, 2][np.newaxis]
            xyb = np.hstack((xy_new, a))
            boundaries = np.insert(boundaries, running_index + 1, xyb, axis=0)

            # 2, because we have to count the new one
            running_index += 2
            index1 += 1
            nr_nodes += 1

            if running_index == index1:
                running_index = index0

    return boundaries


def main():
    options, args = handle_cmd_options()

    # determine a unique directory name
    pwdx = os.getcwd()

    electrodes = np.loadtxt('electrodes.dat')
    boundaries_raw = np.loadtxt('boundaries.dat')

    check_boundaries(boundaries_raw)

    if options.add_stab_nodes:
        boundaries = add_stabilizer_nodes(boundaries_raw,
                                          electrodes,
                                          options.add_stab_nodes)
    else:
        boundaries = boundaries_raw

    # create output directory
    directory = 'tmp_grid_' + str(uuid.uuid4())
    print('Using directory: ' + directory)
    os.makedirs(directory)

    shutil.copy('electrodes.dat', directory + os.sep + 'electrodes.dat')
    shutil.copy('boundaries.dat', directory + os.sep + 'boundaries.dat')

    if os.path.isfile('extra_nodes.dat'):
        shutil.copy('extra_nodes.dat', directory + os.sep + 'extra_nodes.dat')
    if(os.path.isfile('char_length.dat')):
        shutil.copy('char_length.dat', directory + os.sep + 'char_length.dat')
    if os.path.isfile('extra_lines.dat'):
        shutil.copy('extra_lines.dat',
                    directory + os.sep + 'extra_lines.dat')
    if os.path.isfile('gmsh_commands.dat'):
        shutil.copy('gmsh_commands.dat',
                    directory + os.sep + 'gmsh_commands.dat')

    os.chdir(directory)

    mesh = Mesh()
    mesh.read_char_lengths('char_length.dat', 'electrodes.dat')

    # create boundary lines from boundary nodes
    for index in range(0, boundaries.shape[0]):
        print(index, (index + 1) % boundaries.shape[0])
        a = index
        b = (index + 1) % boundaries.shape[0]

        mesh.add_boundary(
            boundaries[a, 0:2], boundaries[b, 0:2], boundaries[a, 2])

    os.makedirs('step1')

    mesh.read_electrodes(electrodes)

    if os.path.isfile('extra_nodes.dat'):
        mesh.read_extra_nodes('extra_nodes.dat')

    if os.path.isfile('extra_lines.dat'):
        mesh.read_extra_lines('extra_lines.dat')

    mesh.write_geo_file('step1/commands.geo')
    mesh.write_electrodes('step1/electrode_positions.dat')
    mesh.write_boundaries('step1/boundary_lines.dat')
    os.chdir('step1')

    gmsh_binary = cBin.get('gmsh')
    print('Calling binary: {0}'.format(gmsh_binary))
    return_code = subprocess.call(
        '{0} -format msh2 -2 commands.geo'.format(gmsh_binary),
        shell=True
    )
    if return_code != 0:
        raise Exception('There was an error during execution of GMSH')
    os.makedirs('step2')
    os.chdir('step2')
    import cr_trig_parse_gmsh
    cr_trig_parse_gmsh.main()
    # return_code = subprocess.call('cr_trig_parse_gmsh.py', shell=True)
    # if return_code != 0:
    #     raise Exception(
    #         'There was an error during execution of cr_trig_parse_gmsh.sh'
    #     )

    #
    os.makedirs('grid')
    shutil.copy('elec.dat', 'grid/elec.dat')
    shutil.copy('../elem.dat', 'grid/elem.dat')
    os.chdir('grid')
    cutmck_binary = cBin.get('CutMcK')
    print('Calling binary: {0}'.format(cutmck_binary))
    return_code = subprocess.call(cutmck_binary, shell=True)
    if return_code != 0:
        raise Exception(
            'There was an error during execution of CutMcK'
        )

    # for convenience, copy the final grid files to the top level
    shutil.copy('elem.dat', '../../../elem.dat')
    shutil.copy('elec.dat', '../../../elec.dat')
    os.chdir(pwdx)

    print('The final grid can be found in:')

    # if we have one parameter, use this as the output directory
    if len(args) == 1:
        output_dir = args[0]
        if not os.path.isdir(output_dir):
            shutil.move(directory, output_dir)
            print(output_dir)
        else:
            print('temporary directory:')
            print(directory)
            raise Exception('output directory already exists!')
    else:
        print(directory)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Parse a GMSH mesh and produce a FEM grid for CRMod/CRTomo. This script should
not be called directly, as it requires the input and output files from
"cr_trig_create.py".

TODO
----

* the program at some point looks for nodes on the boundary line. Perhaps some
  funtions in "grid_extralines_gen_decouplings.py" can be used here (code
  deduplication and speed improvements)


"""
from crtomo.mpl_setup import *
mpl.rcParams['font.size'] = 6.0
import numpy as np


def parse_gmsh(filename, boundary_file):
    """
    Parse a GMSH .msh file and return a dictionary containing the data
    neccessary to create CRTomo grids
    """
    mesh = {}

    fid = open(filename, 'r')
    line = fid.readline()
    while(line):
        if(line.startswith('$MeshFormat')):
            pass
        elif(line.startswith('$Nodes')):
            nodes = []
            line = fid.readline()
            nr_nodes = np.fromstring(line, dtype=int, count=1, sep=r'\n')
            nr_nodes
            while(line):
                line = fid.readline()
                if(line.startswith('$EndNodes')):
                    break
                node = np.fromstring(line, dtype=float, sep=' ')
                nodes.append(node)
            mesh['nodes'] = nodes
        elif(line.startswith('$Elements')):
            """
            Create a dictionary with the element types as keys. E.g.:
            elements['15'] provides all elements of type 15 (Points)
            """
            elements = {}
            line = fid.readline()
            nr_elements = np.fromstring(line, dtype=int, count=1, sep=r'\n')
            nr_elements
            while(line):
                line = fid.readline()
                if(line.startswith('$EndElements')):
                    break
                element = np.fromstring(line, dtype=int, sep=' ')
                # el_nr = element[0]
                el_type = element[1]
                el_nr_tags = element[2]
                # el_tags = element[3:3 + el_nr_tags]
                el_nodes = element[3 + el_nr_tags:]

                # now decide where to put it
                key = str(el_type)
                if(key in elements.keys()):
                    elements[key].append(el_nodes)
                else:
                    elements[key] = []
                    elements[key].append(el_nodes)

            mesh['elements'] = elements
        line = fid.readline()

    fid.close()

    # if boundary_file is != None, then sort the lines (element type 1)
    # according to the element types
    boundaries = {}

    if(boundary_file is not None):
        # load the original boundary lines
        # it is possible that GMSH added additional nodes on these lines, and
        # that is why we need to find all mesh lines that lie on these original
        # lines.
        bids = np.loadtxt(boundary_file)

        for btype in ('12', '11'):
            # select all original boundaries with this type
            a = np.where(bids[:, 4] == int(btype))[0]
            boundaries[btype] = []
            # for each of those lines, find all lines of the mesh that belong
            # here
            for orig_line in bids[a, :]:
                # print('Find all lines lying on the line: ')
                found_one_line = False
                # print(orig_line)
                # construct line equation

                # x1 == x2 ?
                # split into coordinates
                ox1 = orig_line[0]
                ox2 = orig_line[2]
                oy1 = orig_line[1]
                oy2 = orig_line[3]

                if(orig_line[0] == orig_line[2]):
                    # special case: we only need to find all lines with x1 ==
                    # x2 == x1_orig and y_min >= y_orig_min and y_max <=
                    # <_orig_max
                    for line in elements['1']:
                        if(btype == '11'):
                            if(line[0] == 48 and line[1] == 150):
                                pass
                                # print('Find all lines lying on the line: ')
                                # print('This is the line')

                        # it doesn't matter any more to be able to assign x ->
                        # y values. Thus we can sort the y values and just
                        # check
                        # if the new line lies in between the original one
                        oy1, oy2 = np.sort([orig_line[1], orig_line[3]])
                        x1, x2 = np.sort(
                            [
                                mesh['nodes'][line[0] - 1][1],
                                mesh['nodes'][line[1] - 1][1]
                            ]
                        )
                        y1, y2 = np.sort(
                            [
                                mesh['nodes'][line[0] - 1][2],
                                mesh['nodes'][line[1] - 1][2]
                            ]
                        )

                        if np.isclose(x1, x2) and np.isclose(x2, ox1):
                            if(y1 >= oy1 and y2 <= oy2):
                                found_one_line = True
                                boundaries[btype].append(line)

                else:
                    # print('checking with full line equation')
                    # no vertical line
                    # we need the full check using the line equation
                    slope = (orig_line[1] - orig_line[3]) / (
                        orig_line[0] - orig_line[2])
                    intersect = orig_line[1] - (slope * orig_line[0])
                    # print('Slope', slope, ' Intercept ', intersect)
                    for line in elements['1']:
                        x1 = mesh['nodes'][line[0] - 1][1]
                        y1 = mesh['nodes'][line[0] - 1][2]
                        x2 = mesh['nodes'][line[1] - 1][1]
                        y2 = mesh['nodes'][line[1] - 1][2]

                        # print(x1, x2, y1, y1)
                        check = False
                        # check if x coordinates of the test line fit in the
                        # original line
                        if(ox1 < ox2):
                            if(x1 < x2):
                                if((np.isclose(x1, ox1) or x1 > ox1) and
                                   (np.isclose(x2, ox2) or x2 < ox2)):
                                    check = True
                            else:
                                if((np.isclose(x2, ox1) or x2 >= ox1) and
                                   (np.isclose(x1, ox2) or x1 <= ox2)):
                                    check = True
                        else:
                            if(x1 < x2):
                                if((np.isclose(x1, ox2) or x1 >= ox2) and
                                   (np.isclose(x2, ox1) or x2 <= ox1)):
                                    check = True
                            else:
                                if((np.isclose(x2, ox2) or x2 >= ox2) and
                                   (np.isclose(x1, ox1) or x1 <= ox1)):
                                    check = True

                        # print('boundary check:', check)
                        if(check):
                            # the line lies within the x-range of the orig line
                            ytest1 = slope * x1 + intersect
                            ytest2 = slope * x2 + intersect
                            if(np.around(ytest1 - y1, 5) == 0 and
                               np.around(ytest2 - y2, 5) == 0):
                                boundaries[btype].append(line)
                                # found = True
                                found_one_line = True
                                # print('found it new', line)
                # add a weak check: we need to find at least one line in the
                # mesh corresponding to this boundary line:
                if not found_one_line:
                    raise Exception('no mesh line found for this boundary')

            print('Total number of boundaries of this type:',
                  len(boundaries[btype]))
    mesh['boundaries'] = boundaries
    return mesh


class _line():
    def __init__(self, p1, p2):
        self.p1_x = p1[0]
        self.p1_y = p1[1]

        self.p2_x = p2[0]
        self.p2_y = p2[1]

    def get_diff(self):
        diff = []
        diff.append(self.p1_x - self.p2_x)
        diff.append(self.p1_y - self.p2_y)
        return np.array(diff)

    def get_distance(self):
        diff = self.get_diff()
        dist = np.sqrt(np.sum(diff ** 2))
        return dist

    def get_center(self):
        """
        Return (x,y) coordinate of central point
        """
        dist = self.get_distance()
        diff = self.get_diff()
        direction = diff / dist

        center_x = self.p2_x + (dist * 0.5) * direction[0]
        center_y = self.p2_y + (dist * 0.5) * direction[1]

        return np.array([center_x, center_y])


def debug_plot_mesh(mesh, boundary_elements):
    plot_large = True

    # prepare nodes
    nodes = np.array(mesh['nodes'])
    tx = nodes[:, 1]
    ty = nodes[:, 2]

    # adapt height of plot to size of grid
    tx_size = np.abs(np.max(tx) - np.min(tx))
    ty_size = np.abs(np.max(ty) - np.min(ty))

    if(plot_large):
        width = 10
    else:
        width = 7
    size_x = width
    size_y = width * (ty_size / tx_size) * 1.5

    # plot triangles
    triangles = np.array(mesh['elements']['2']) - 1

    fig, ax = plt.subplots(1, 1, figsize=(size_x, size_y))
    ax.triplot(tx, ty, triangles, color='k', linewidth=1.0)

    # plot boundaries in red
    # lines = np.array(mesh['elements']['1']) - 1
    lines = np.array(mesh['boundaries']['12'] + mesh['boundaries']['11']) - 1
    lx = nodes[lines, 1]
    ly = nodes[lines, 2]

    for index in range(0, lx.shape[0]):
        # for index in range(0, 1):
        ax.plot(lx[index, :], ly[index, :], 'r.', alpha=0.4)
        # draw a line to the neighbour
        line_obj = _line([lx[index, 0], ly[index, 0]],
                         [lx[index, 1], ly[index, 1]])
        center = line_obj.get_center()
        ntri = np.array(mesh['elements']['2'])[int(boundary_elements[index])]
        ntri_x = nodes[ntri - 1, 1]
        ntri_y = nodes[ntri - 1, 2]

        # compute centroid of element adjacent to the boundary element
        centroidx = np.sum(ntri_x) / 3
        centroidy = np.sum(ntri_y) / 3

        ax.plot([centroidx, center[0]], [centroidy, center[1]], color='g',
                label='adjacent elements')
        ax.annotate('{0}'.format(index), xy=(centroidx, centroidy),
                    fontsize=6.0, color='k')

    ax.set_xlim([np.min(tx) - 0.1, np.max(tx) + 0.1])
    ax.set_ylim([np.min(ty) - 0.1, np.max(ty) + 0.1])
    ax.set_aspect('equal')

    # plot electrodes
    electrodes = np.loadtxt('../electrode_positions.dat')
    ax.scatter(electrodes[:, 0], electrodes[:, 1], s=30,
               color='blue',
               label='electrodes')

    fig.tight_layout()
    if(plot_large):
        dpi = 600
    else:
        dpi = 300
    fig.savefig('../../triangle_grid.jpg', dpi=dpi)


def get_header(mesh):
    """
    For now we only know Element types 1 (line) and 2 (triangle)
    """
    nr_all_nodes = len(mesh['nodes'])
    nr_types = 1  # triangles

    # compute bandwidth
    bandwidth = -1
    for triangle in mesh['elements']['2']:
        diff1 = abs(triangle[0] - triangle[1])
        diff2 = abs(triangle[0] - triangle[2])
        diff3 = abs(triangle[1] - triangle[2])
        diffm = max(diff1, diff2, diff3)
        if(diffm > bandwidth):
            bandwidth = diffm

    el_infos = []
    # triangles
    for element in ('2',):
        el = mesh['elements'][element]
        if(element == '2'):
            el_type = 3
        elif(element == '1'):
            el_type = 12  # Neumann
        nr = len(el)
        nr_nodes = len(el[0])
        el_infos.append((el_type, nr, nr_nodes))

    # boundary elements
    for btype in ('12', '11'):
        if(btype in mesh['boundaries']):
            el_type = int(btype)
            nr = len(mesh['boundaries'][btype])
            nr_nodes = 2
            if(nr > 0):
                nr_types += 1
                el_infos.append((el_type, nr, nr_nodes))

    # now convert to string
    str_header = ''
    for a, b, c in [(nr_all_nodes, nr_types, bandwidth), ] + el_infos:
        str_header = str_header + '{0}  {1} {2}\n'.format(a, b, c)
    return str_header


def get_nodes(mesh):
    nodes = np.array(mesh['nodes'])

    str_nodes = ''
    for node_nr in range(0, nodes.shape[0]):
        str_nodes += '{0}   {1} {2}\n'.format(node_nr + 1, nodes[node_nr, 1],
                                              nodes[node_nr, 2])
    return str_nodes


def get_elements(mesh):
    # print('Get elements')
    str_elements = ''
    # triangles
    for element in ('2'):
        el = mesh['elements'][element]
        for item in el:
            # the reverse should ensure counter-clockwise element nodes
            if(element == '1'):
                # boundary elements
                lst = item
            else:
                lst = reversed(item)

            for i in lst:
                str_elements += '{0}  '.format(i)
            str_elements += '\n'
    # boundary elements
    for btype in ('12', '11'):
        for line in mesh['boundaries'][btype]:
            str_elements += '{0} {1}\n'.format(line[0], line[1])

    return(str_elements)


def get_ajd_bound(mesh):
    """
    Determine triangular elements adjacend to the boundary elements
    """
    print('Get elements adjacent to boundaries')
    boundary_elements = []
    str_adj_boundaries = ''
    # for boundary in mesh['elements']['1']:
    boundaries = mesh['boundaries']['12'] + mesh['boundaries']['11']
    for boundary in boundaries:
        # now find the triangle ('2') with two nodes equal to this boundary
        indices = [nr if (boundary[0] in x and boundary[1] in x) else np.nan
                   for (nr, x) in enumerate(mesh['elements']['2'])]
        indices = np.array(indices)[~np.isnan(indices)]
        if(len(indices) != 1):
            print('More than one neighbour found!')
        elif(len(indices) == 0):
            print('No neighbour found!')
        boundary_elements.append(indices[0])
        str_adj_boundaries += '{0}\n'.format(int(indices[0]) + 1)
    return str_adj_boundaries, boundary_elements


def write_elec_file(filename, mesh):
    """
    Read in the electrode positions and return the indices of the electrodes

    # TODO: Check if you find all electrodes
    """
    elecs = []
    # print('Write electrodes')
    electrodes = np.loadtxt(filename)
    for i in electrodes:
        # find
        for nr, j in enumerate(mesh['nodes']):
            if np.isclose(j[1], i[0]) and np.isclose(j[2], i[1]):
                elecs.append(nr + 1)

    fid = open('elec.dat', 'w')
    fid.write('{0}\n'.format(len(elecs)))
    for i in elecs:
        fid.write('{0}\n'.format(i))
    fid.close()


def main():
    mesh = parse_gmsh('../commands.msh', '../boundary_lines.dat')

    # now create the CRTomo grid
    """
    1. Header
    2. Nodes
    3. Elements: Triangles, Boundary elements
    4. Element ids for adjoining boundary elements
    """

    str_header = get_header(mesh)
    str_nodes = get_nodes(mesh)
    str_elements = get_elements(mesh)
    str_adj_boundaries, boundary_elements = get_ajd_bound(mesh)

    crt_mesh = str_header + str_nodes + str_elements + str_adj_boundaries

    fid = open('../elem.dat', 'w')
    fid.write(crt_mesh)
    fid.close()

    write_elec_file('../electrode_positions.dat', mesh)
    debug_plot_mesh(mesh, boundary_elements)



"""Representations of CRMod and CRTomo configurations.

Examples
--------

    >>> import crtomo.cfg as CRcfg
        crmod_cfg = CRcfg.crmod_config()
        print(crmod_cfg)
    ***FILES***       !  mswitch
    ../grid/elem.dat       !  elem
    ../grid/elec.dat       !  elec
    ../rho/rho.dat       !  rho
    ../config/config.dat       !  config
    F       !  write_pot
    ../mod/pot/pot.dat       !  pot_file
    T       !  write_volts
    ../mod/volt.dat       !  volt_file
    F       !  write_sens
    ../mod/sens/sens.dat       !  sens_file
    F       !  another_dataset
    1       !  2D
    F       !  fictitious_sink
    1660       !  sink_node
    F       !  boundary_values
    boundary.dat       !  boundary_file

TODO
----

* we could also add help texts here for each parameter

"""


class crmod_config(dict):
    """
    Write CRMod configuration files (crmod.cfg).

    This class is essentially a dict of CRMod configurations with a few extra
    functions that know how to write a proper crmod.cfg file.

    Examples
    --------

    >>> import crtomo.cfg as cCFG
        crmod_cfg = cCfg.crmod_config()
        crmod_cfg.write_to_file('crmod.cfg')

    """
    def __init__(self, *arg, **kw):
        super(crmod_config, self).__init__(*arg, **kw)
        self.set_defaults()

        self.key_order = (
            'mswitch',
            'elem',
            'elec',
            'rho',
            'config',
            'write_pots',
            'pot_file',
            'write_volts',
            'volt_file',
            'write_sens',
            'sens_file',
            'another_dataset',
            '2D',
            'fictitious_sink',
            'sink_node',
            'boundary_values',
            'boundary_file'
        )

        # boolean options
        self.bools = (
            'write_pots',
            'write_volts',
            'write_sens',
        )

    def _check_and_convert_bools(self):
        """Replace boolean variables by the characters 'F'/'T'
        """
        replacements = {
            True: 'T',
            False: 'F',
        }

        for key in self.bools:
            if isinstance(self[key], bool):
                self[key] = replacements[self[key]]

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        new_copy = crmod_config()
        # translate the keys
        for key in self.keys():
            new_copy[key] = self[key]
        return new_copy

    def __deepcopy__(self, memo):
        print('deepcopy')
        raise Exception('not implemented')

    def set_defaults(self):
        """
        Fill the dictionary with all defaults
        """
        self['mswitch'] = '***FILES***'
        self['elem'] = '../grid/elem.dat'
        self['elec'] = '../grid/elec.dat'
        self['rho'] = '../rho/rho.dat'
        self['config'] = '../config/config.dat'
        self['write_pots'] = 'F'  # ! potentials ?
        self['pot_file'] = '../mod/pot/pot.dat'
        self['write_volts'] = 'T'  # ! measurements ?
        self['volt_file'] = '../mod/volt.dat'
        self['write_sens'] = 'F'  # ! sensitivities ?
        self['sens_file'] = '../mod/sens/sens.dat'
        self['another_dataset'] = 'F'  # ! another dataset ?
        self['2D'] = '1'  # ! 2D (=0) or 2.5D (=1)
        self['fictitious_sink'] = 'F'  # ! fictitious sink ?
        self['sink_node'] = '1660'  # ! fictitious sink node number
        self['boundary_values'] = 'F'  # ! boundary values ?
        self['boundary_file'] = 'boundary.dat'

    def write_to_file(self, filename):
        """
        Write the configuration to a file. Use the correct order of values.
        """
        self._check_and_convert_bools()
        fid = open(filename, 'w')

        for key in self.key_order:
            if(key == -1):
                fid.write('\n')
            else:
                fid.write('{0}\n'.format(self[key]))

        fid.close()

    def __repr__(self):
        self._check_and_convert_bools()
        representation = ''
        for key in self.key_order:
            if key == -1:
                representation += '\n'
            else:
                representation += '{0}       !  {1}\n'.format(self[key], key)
        return representation

    def __str__(self):
        return self.__repr__()


class crtomo_config(dict):
    """
    Write CRTomo configuration files (crtomo.cfg).

    This class is essentially a dict of CRTomo configurations with a few extra
    functions that know how to write a proper crtomo.cfg file.
    """
    def __init__(self, *arg, **kw):
        super(crtomo_config, self).__init__(*arg, **kw)
        self.set_defaults()

        # -1 indicates an empty line
        self.key_order = (
            'mswitch', 'elem', 'elec', 'volt', 'inv_dir', 'diff_inv',
            -1, 'prior_model', -1, 'iseed_var', 'cells_x', 'cells_z',
            'ani_x', 'ani_z', 'max_it', 'dc_inv', 'robust_inv',
            'fpi_inv', 'mag_rel', 'mag_abs', 'pha_a1', 'pha_b',
            'pha_rel', 'pha_abs', 'hom_bg', 'hom_mag', 'hom_pha',
            'another_ds', 'd2_5', 'fic_sink', 'fic_sink_node',
            'boundaries', 'boundaries_file', 'mswitch2', 'lambda',
        )

        self.mswitch_values = {
            'l1_cov_dw': 1,
            'lcov1': 2,
            'res_m': 4,
            'lcov2': 8,
            'ols_gauss': 16,
            'err_ellipse': 32,
            'force_neg_phase': 128,
            'lsytop': 256,
            'lvario': 512,
            'verbose': 1024,
            'verbose_dat': 2048,
        }

    def set_defaults(self):
        """Fill the dictionary with all defaults
        """
        self['mswitch'] = 1
        self['elem'] = '../grid/elem.dat'
        self['elec'] = '../grid/elec.dat'
        self['volt'] = '../mod/volt.dat'
        self['inv_dir'] = '../inv'
        self['prior_model'] = ''
        self['diff_inv'] = 'F ! difference inversion?'
        self['iseed_var'] = 'iseed variance'
        self['cells_x'] = '0    ! # cells in x-direction'
        self['cells_z'] = '0    ! # cells in z-direction'
        self['ani_x'] = '1.000  ! smoothing parameter in x-direction'
        self['ani_z'] = '1.000  ! smoothing parameter in z-direction'
        self['max_it'] = '20    ! max. nr of iterations'
        self['dc_inv'] = 'F     ! DC inversion?'
        self['robust_inv'] = 'T     ! robust inversion?'
        self['fpi_inv'] = 'F     ! final phase improvement?'
        self['mag_rel'] = '5'
        self['mag_abs'] = '1e-3'
        self['pha_a1'] = 0
        self['pha_b'] = 0
        self['pha_rel'] = 0
        self['pha_abs'] = 0
        self['hom_bg'] = 'F'
        self['hom_mag'] = '10.00'
        self['hom_pha'] = '0.00'
        self['another_ds'] = 'F'
        self['d2_5'] = '1'
        self['fic_sink'] = 'F'
        self['fic_sink_node'] = '10000'
        self['boundaries'] = 'F'
        self['boundaries_file'] = 'boundary.dat'
        self['mswitch2'] = '1'
        self['lambda'] = 'lambda'

    def set_mswitch(self, key, active):
        """The mswitch can enable/disable various functions of CRTomo, mainly
        concerned with the computation of various resolution parameters.
        The switch itself is implemented as a binary switch, meaning that each
        function can be enabled/disabled by setting its corresponding bit.

        This function simplifies control of these functions by providing a
        simple boolean interface for these options.

        Possible keys: {}

        Parameters
        ----------
        key : str
            Function to control
        active : bool
            Activate (True) or deactivate (False) the feature
        """.format(
            ['{}'.format(x) for x in self.mswitch_values.keys()]
        )

        assert key in self.mswitch_values
        if active:
            self['mswitch'] = (
                self['mswitch'] % self.mswitch_values[key]
            ) + self.mswitch_values[key]
        else:
            self['mswitch'] = (self['mswitch'] % self.mswitch_values[key])

    def write_to_file(self, filename):
        """ Write the configuration to a file. Use the correct order of values.
        """
        fid = open(filename, 'w')

        for key in self.key_order:
            if(key == -1):
                fid.write('\n')
            else:
                fid.write('{0}\n'.format(self[key]))

        fid.close()

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        new_copy = crtomo_config()
        # translate the keys
        for key in self.keys():
            new_copy[key] = self[key]
        return new_copy

    def __deepcopy__(self, memo):
        print('deepcopy')
        raise Exception('not implemented')

    def __repr__(self):
        representation = ''
        for key in self.key_order:
            if key == -1:
                representation += '\n'
            else:
                representation += '{0}       !  {1}\n'.format(self[key], key)
        return representation

    def __str__(self):
        return self.__repr__()

    def help(key):
        """Return the help text specific to a certain key
        """
        help_dict = {

        }
        return_text = help_dict.get(key, 'no help available')
        return return_text

    def import_from_file(self, filename):
        """Import a CRTomo configuration from an existing crtomo.cfg file

        Parameters
        ----------
        filename : str
            Path to crtomo.cfg file
        """
        lines_raw = [x.strip() for x in open(filename, 'r').readlines()]
        key_index = 0
        for line in lines_raw:
            # ignore comments
            if line.startswith('#'):
                continue
            else:
                # check if we have inline comments
                index_inline_comment = line.find('!')
                if index_inline_comment != -1:
                    line = line[0:index_inline_comment].strip()
            key = self.key_order[key_index]
            if key == -1:
                pass
            else:
                self[key] = line
            # increment the key_index
            key_index += 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Clean a simulation directory of all modeling/inversion files
"""
import numpy as np
import os
import glob


def main():
    rm_list = []

    required_files_inversion = (
        'exe/crtomo.cfg',
        'grid/elem.dat',
        'grid/elec.dat',
        'mod/volt.dat')
    clean_inv = np.all([os.path.isfile(x) for x in required_files_inversion])
    if clean_inv:
        rm_list += glob.glob('inv/*')
        rm_list += [
            'exe/error.dat',
            'exe/crtomo.pid',
            'exe/variogram.gnu',
            'exe/inv.elecpositions',
            'exe/inv.gstat',
            'exe/inv.lastmod',
            'exe/inv.lastmod_rho',
            'exe/inv.mynoise_pha',
            'exe/inv.mynoise_rho',
            'exe/inv.mynoise_voltages',
            'exe/tmp.kfak',
            'overview.png',
        ]

    required_files_modelling = (
        'exe/crmod.cfg',
        'grid/elem.dat',
        'grid/elec.dat',
        'config/config.dat',
        'rho/rho.dat'
    )
    clean_mod = np.all([os.path.isfile(x) for x in required_files_modelling])
    if clean_mod:
        rm_list += glob.glob('mod/sens/*')
        rm_list += glob.glob('mod/pot/*')
        rm_list += ['mod/volt.dat', ]
        rm_list += ['exe/crmod.pid', ]

    for filename in rm_list:
        if os.path.isfile(filename):
            # print('Removing file {0}'.format(filename))
            os.remove(filename)

    plot_files = (
        'rho.png',
        'imag.png',
        'real.png',
        'phi.png',
        'cov.png',
        'fpi_imag.png',
        'fpi_phi.png',
        'fpi_real.png',
    )
    for filename in plot_files:
        if os.path.isfile(filename):
            os.remove(filename)


if __name__ == '__main__':
    main()



"""Manage and plot CRMod/CRTomo grids

Creating grids
--------------

Grids can be created in various ways:

    * Regular grids can be created using Griev (which is not supported any
      more)
    * Irregular (triangular) grids are created using the command line tool
      'cr_trig_create'
    * the crt_grid class provides some simple wrapper functions for
      cr_trig_create that directory load created grids

Notes
-----

After loading an elem.dat file, i.e., the grid, grid information are exposed
via various structures:

    * the header is stored in the dict self.header::

          >>> self.header.keys()
              dict_keys(
                ['bandwidth', 'element_infos', 'cutmck', 'nr_element_types',
                'nr_nodes']
              )

    * nodes are stored in self.nodes. Various sortings are available. If in
      doubt use self.nodes['presort'].
    * elements are stored in self.elements (as node numbers)


Examples
--------

>>> import crtomo.grid as CRTGrid
    grid = CRTGrid.crt_grid()
    grid.load_grid('elem.dat', 'elec.dat')

>>> # extracting element coordinates in the order of rho.dat file
    import crtomo.grid as CRGrid
    grid = CRGrid.crt_grid(elem_file='elem.dat', elec_file='elec.dat')
    # 1. element (oben links)
    grid.nodes['presort'][grid.elements[0], :]
    # columns: node number - x position - z position
"""
import tempfile
import subprocess
import os
import time

import numpy as np
import scipy.sparse
from scipy.spatial.distance import pdist
import scipy.spatial
import pandas as pd

import crtomo.mpl
plt, mpl = crtomo.mpl.setup()


class crt_grid(object):
    """The :class:`crt_grid` holds and manages one Finite-Element grid.

    This class provides only basic plotting routines. Advanced plotting
    routines can be found in :class:`crtomo.plotManager`.

    Examples
    --------

    >>> import crtomo.grid as CRGrid
        grid = CRGrid(elem_file='elem.dat', elec_file='elec.dat')
        print(grid)
    CRMod/CTRomo grid instance
    number of elements: 2728
    number of nodes: 1440
    number of electrodes: 60
    grid dimsensions:
        X: -7.5 37.5
        Z: -15.0 0.0


    """
    def __init__(self, elem_file=None, elec_file=None, empty=False):
        """

        Parameters
        ----------


        """
        self.electrodes = None
        self.header = None
        self.nodes = None
        self.elements = None
        # store the neighbors (by zeroed index) of each element
        self.element_neighbors_data = None
        # for each neighbor, store the nodes common to both elements
        self.element_neighbors_edges = None
        self.grid_is_rectangular = None
        self.grid_data = []

        self.props = {
            'electrode_color': 'k',
        }
        self.grid = None
        self.nr_of_elements = None
        self.nr_of_nodes = None
        self.nr_of_electrodes = None

        # will be used for caching by .get_element_centroids
        self.centroids = None

        if elem_file is None and not empty:
            # check if elem.dat is present
            if os.path.isfile('elem.dat'):
                elem_file = 'elem.dat'

        if elec_file is None and not empty:
            # check if elec.dat is present
            if os.path.isfile('elec.dat'):
                elec_file = 'elec.dat'

        if elem_file is not None:
            self.load_elem_file(elem_file)

        if elec_file is not None:
            self.load_elec_file(elec_file)

    def __repr__(self):
        output_str = 'CRMod/CTRomo grid instance\n'
        output_str += 'number of elements: {0}\n'.format(self.nr_of_elements)
        output_str += 'number of nodes: {0}\n'.format(self.nr_of_nodes)
        output_str += 'number of electrodes: {0}\n'.format(
            self.nr_of_electrodes
        )
        output_str += 'grid dimsensions: \n'
        xlim, zlim = self.get_minmax()
        output_str += 'X: {0} {1} \n'.format(*xlim)
        output_str += 'Z: {0} {1} \n'.format(*zlim)

        return output_str

    class element:
        def __init__(self):
            self.nodes = np.empty(1)
            self.xcoords = np.empty(1)
            self.zcoords = np.empty(1)

        def center(self):
            xcenter = np.mean(self.xcoords)
            zcenter = np.mean(self.zcoords)
            return (xcenter, zcenter)

        def length_line(self):
            assert len(self.xcoords) == 2
            length = np.sqrt(
                np.sum(
                    np.diff(self.xcoords) ** 2 + np.diff(self.zcoords) ** 2
                )
            )
            return length

        def vector_line(self):
            """Return the vector of the boundary element"""
            assert len(self.xcoords) == 2
            diff_x = self.xcoords[1] - self.xcoords[0]
            diff_z = self.zcoords[1] - self.zcoords[0]
            vec = np.hstack((diff_x, diff_z))
            return vec

        def vector_norm_line(self):
            return self.vector_line() / self.length_line()

        def outer_normal_line(self):
            vector = self.vector_norm_line()
            # element coords are provided counter-clockwise.
            # for outer normal, we need to rotate by 90 degree to the left
            vector[1] *= -1
            # swap x and z coords
            return vector[::-1]

    def add_grid_data(self, data):
        """ Return id
        """
        self.grid_data.append(data)
        return len(self.grid_data) - 1

    def _read_elem_header(self, fid):
        header = {}
        # first header line
        firstline = fid.readline().lstrip()
        nr_of_nodes, nr_of_element_types, bandwidth = np.fromstring(
            firstline, dtype=int, sep='   ')
        header['nr_nodes'] = nr_of_nodes
        header['nr_element_types'] = nr_of_element_types
        header['bandwidth'] = bandwidth

        # read header lines for each element type
        element_infos = np.zeros((nr_of_element_types, 3), dtype=int)
        for element_type in range(0, nr_of_element_types):
            element_line = fid.readline().lstrip()
            element_infos[element_type, :] = np.fromstring(
                element_line,
                dtype=int,
                sep='  ',
            )
        header['element_infos'] = element_infos
        self.header = header

    def _read_elem_nodes(self, fid):
        """ Read the nodes from an opened elem.dat file. Correct for CutMcK
        transformations.

        We store three typed of nodes in the dict 'nodes':

            * "raw" : as read from the elem.dat file
            * "presort" : pre-sorted so we can directly read node numbers from
              a elec.dat file and use them as indices.
            * "sorted" : completely sorted as in the original grid (before any
                       CutMcK)

        For completeness, we also store the following keys:

            * "cutmck_index" : Array containing the indices in "presort" to
              obtain the "sorted" values:
              nodes['sorted'] = nodes['presort'] [nodes['cutmck_index'], :]
            * "rev_cutmck_index" : argsort(cutmck_index)
        """
        nodes = {}

        #   # prepare nodes
        #   nodes_sorted = np.zeros((number_of_nodes, 3), dtype=float)
        #   nodes = np.zeros((number_of_nodes, 3), dtype=float)

        # read in nodes
        nodes_raw = np.empty((self.header['nr_nodes'], 3), dtype=float)
        for nr in range(0, self.header['nr_nodes']):
            node_line = fid.readline().lstrip()
            nodes_raw[nr, :] = np.fromstring(
                node_line, dtype=float, sep='    ')

        # round node coordinates to 5th decimal point. Sometimes this is
        # important when we deal with mal-formatted node data
        nodes_raw[:, 1:3] = np.round(nodes_raw[:, 1:3], 5)

        # check for CutMcK
        # The check is based on the first node, but if one node was renumbered,
        # so were all the others.
        if(nodes_raw[:, 0] != list(range(1, nodes_raw.shape[0]))):
            self.header['cutmck'] = True
            print(
                'This grid was sorted using CutMcK. The nodes were resorted!')
        else:
            self.header['cutmck'] = False

        # Rearrange nodes when CutMcK was used.
        if(self.header['cutmck']):
            nodes_cutmck = np.empty_like(nodes_raw)
            nodes_cutmck_index = np.zeros(nodes_raw.shape[0], dtype=int)
            for node in range(0, self.header['nr_nodes']):
                new_index = np.where(nodes_raw[:, 0].astype(int) == (node + 1))
                nodes_cutmck[new_index[0], 1:3] = nodes_raw[node, 1:3]
                nodes_cutmck[new_index[0], 0] = new_index[0]
                nodes_cutmck_index[node] = new_index[0]
            # sort them
            nodes_sorted = nodes_cutmck[nodes_cutmck_index, :]
            nodes['presort'] = nodes_cutmck
            nodes['cutmck_index'] = nodes_cutmck_index
            nodes['rev_cutmck_index'] = np.argsort(nodes_cutmck_index)
        else:
            nodes_sorted = nodes_raw
            nodes['presort'] = nodes_raw

        # prepare node dict
        nodes['raw'] = nodes_raw
        nodes['sorted'] = nodes_sorted

        self.nodes = nodes
        self.nr_of_nodes = nodes['raw'].shape[0]

    def _read_elem_elements(self, fid):
        """Read all FE elements from the file stream. Elements are stored in
        the self.element_data dict. The keys refer to the element types:

         *  3: Triangular grid (three nodes)
         *  8: Quadrangular grid (four nodes)
         * 11: Mixed boundary element
         * 12: Neumann (no-flow) boundary element

        """
        elements = {}

        # read elements
        for element_type in range(0, self.header['nr_element_types']):
            element_list = []
            for element_coordinates in range(
                    0, self.header['element_infos'][element_type, 1]):
                element_coordinates_line = fid.readline().lstrip()
                tmp_element = self.element()
                tmp_element.nodes = np.fromstring(element_coordinates_line,
                                                  dtype=int, sep=' ')
                tmp_element.xcoords = self.nodes['presort'][tmp_element.nodes -
                                                            1, 1]
                tmp_element.zcoords = self.nodes['presort'][tmp_element.nodes -
                                                            1, 2]
                element_list.append(tmp_element)
            element_type_number = self.header['element_infos'][element_type, 0]
            elements[element_type_number] = element_list
        self.element_data = elements

    def _prepare_grids(self):
        """
        depending on the type of grid (rectangular or triangle), prepare grids
        or triangle lists

        TODO: We want some nice way of not needing to know in the future if we
              loaded triangles or quadratic elements.
        """
        if(self.header['element_infos'][0, 2] == 3):
            print('Triangular grid found')
            self.grid_is_rectangular = False

            triangles = self.element_data[3]
            triangles = [x.nodes for x in triangles]
            # python starts arrays with 0, but elem.dat with 1
            triangles = np.array(triangles) - 1
            self.elements = triangles
            tri_x = self.nodes['presort'][triangles, 1]
            tri_z = self.nodes['presort'][triangles, 2]
            self.grid = {}
            self.grid['x'] = tri_x
            self.grid['z'] = tri_z

        else:
            print('Rectangular grid found')
            self.grid_is_rectangular = True
            quads_raw = [x.nodes for x in self.element_data[8]]
            quads = np.array(quads_raw) - 1
            self.elements = quads
            quads_x = self.nodes['presort'][quads, 1]
            quads_z = self.nodes['presort'][quads, 2]
            self.grid = {}
            self.grid['x'] = quads_x
            self.grid['z'] = quads_z

            # calculate the dimensions of the grid
            try:
                self.calculate_dimensions()
            except Exception as e:
                e
                self.nr_nodes_x = None
                self.nr_nodes_z = None
                self.nr_elements_x = None
                self.nr_elements_z = None
        self.nr_of_elements = self.grid['x'].shape[0]

    def calculate_dimensions(self):
        """For a regular grid, calculate the element and node dimensions
        """
        x_coordinates = np.sort(self.grid['x'][:, 0])  # first x node
        self.nr_nodes_z = np.where(x_coordinates == x_coordinates[0])[0].size
        self.nr_elements_x = self.elements.shape[0] / (self.nr_nodes_z - 1)
        self.nr_nodes_x = self.nr_elements_x + 1
        self.nr_elements_z = self.nr_nodes_z - 1

    def get_minmax(self):
        """Return min/max x/z coordinates of grid

        Returns
        -------
        x: [float, float]
            min, max values of grid dimensions in x direction (sideways)
        z: [float, float]
            min, max values of grid dimensions in z direction (downwards)

        """
        x_minmax = [np.min(self.grid['x']), np.max(self.grid['x'].max())]
        z_minmax = [np.min(self.grid['z']), np.max(self.grid['z'].max())]
        return x_minmax, z_minmax

    def _read_elem_neighbors(self, fid):
        """Read the boundary-element-neighbors from the end of the file
        """
        # get number of boundary elements
        # types 11 and 12 are boundary elements
        self.neighbors = {}
        try:
            for key in (11, 12):
                if key not in self.element_data.keys():
                    continue
                length = len(self.element_data[key])
                self.neighbors[key] = []
                for i in range(length):
                    self.neighbors[key].append(
                        int(fid.readline().strip())
                    )
        except Exception as e:
            e
            raise Exception('Not enough neighbors in file')

    def load_elem_file(self, elem_file):
        """

        """
        with open(elem_file, 'r') as fid:
            self._read_elem_header(fid)
            self._read_elem_nodes(fid)
            self._read_elem_elements(fid)
            self._read_elem_neighbors(fid)
        self._prepare_grids()

    def save_elem_file(self, output):
        """Save elem.dat to file.
        The grid is saved as read in, i.e., with or without applied cutmck.
        If you want to change node coordinates, use self.nodes['raw']

        Parameters
        ----------
        filename: string
            output filename
        """
        with open(output, 'wb') as fid:
            self._write_elem_header(fid)
            self._write_nodes(fid)
            self._write_elements(fid)
            self._write_neighbors(fid)

    def save_elec_file(self, filename):
        with open(filename, 'wb') as fid:
            fid.write(
                bytes(
                    '{0}\n'.format(int(self.electrodes.shape[0])),
                    'utf-8',
                )
            )
            # the + 1 fixes the zero-indexing
            np.savetxt(fid, self.electrodes[:, 0].astype(int) + 1, fmt='%i')

    def _write_neighbors(self, fid):
        for key in (11, 12):
            if key in self.neighbors:
                np.savetxt(fid, self.neighbors[key], fmt='%i')

    def _write_elements(self, fid):
        for dtype in self.header['element_infos'][:, 0]:
            for elm in self.element_data[dtype]:
                np.savetxt(fid, np.atleast_2d(elm.nodes), fmt='%i')

    def _write_nodes(self, fid):
        np.savetxt(fid, self.nodes['raw'], fmt='%i %f %f')

    def _write_elem_header(self, fid):
        fid.write(
            bytes(
                '{0} {1} {2}\n'.format(
                    self.header['nr_nodes'],
                    self.header['nr_element_types'],
                    self.header['bandwidth']),
                'utf-8',
            )
        )
        np.savetxt(fid, self.header['element_infos'], fmt='%i')

    def load_elec_file(self, elec_file):
        electrode_nodes_raw = np.loadtxt(elec_file, skiprows=1, dtype=int) - 1
        self.electrodes = self.nodes['presort'][electrode_nodes_raw]
        self.nr_of_electrodes = self.electrodes.shape[0]

    def load_grid(self, elem_file, elec_file):
        """Load elem.dat and elec.dat
        """
        self.load_elem_file(elem_file)
        self.load_elec_file(elec_file)

    def get_electrode_node(self, electrode):
        """
        For a given electrode (e.g. from a config.dat file), return the true
        node number as in self.nodes['sorted']
        """
        elec_node_raw = int(self.electrodes[electrode - 1][0])
        if(self.header['cutmck']):
            elec_node = self.nodes['rev_cutmck_index'][elec_node_raw]
        else:
            elec_node = elec_node_raw - 1
        return int(elec_node)

    def plot_grid_to_ax(self, ax, **kwargs):
        """
        Other Parameters
        ---------------
        plot_electrode_numbers: bool, optional
            Plot electrode numbers in the grid, default: False
        """
        all_xz = []
        for x, z in zip(self.grid['x'], self.grid['z']):
            tmp = np.vstack((x, z)).T
            all_xz.append(tmp)
        collection = mpl.collections.PolyCollection(
            all_xz,
            edgecolor='k',
            facecolor='none',
            linewidth=0.4,
        )
        ax.add_collection(collection)
        if self.electrodes is not None:
            ax.scatter(
                self.electrodes[:, 1],
                self.electrodes[:, 2],
                color=self.props['electrode_color'],
                clip_on=False,
            )
        ax.set_xlim(self.grid['x'].min(), self.grid['x'].max())
        ax.set_ylim(self.grid['z'].min(), self.grid['z'].max())
        # ax.autoscale_view()
        ax.set_aspect('equal')

        if kwargs.get('plot_electrode_numbers', False):
            for nr, xy in enumerate(self.electrodes[:, 1:3]):
                ax.text(
                    xy[0], xy[1],
                    format(nr + 1),
                    bbox=dict(boxstyle='circle', facecolor='red', alpha=0.8)
                )

    def plot_grid(self, **kwargs):
        """
        Other Parameters
        ----------------
        plot_electrode_numbers: bool, optional
            Plot electrode numbers in the grid, default: False
        """
        fig, ax = plt.subplots(1, 1)
        self.plot_grid_to_ax(ax, **kwargs)
        return fig, ax

    def test_plot(self):
        # play with plot routines
        fig, ax = plt.subplots(1, 1)
        all_xz = []
        for x, z in zip(self.grid['x'], self.grid['z']):
            tmp = np.vstack((x, z)).T
            all_xz.append(tmp)
        collection = mpl.collections.PolyCollection(all_xz, edgecolor='r')
        ax.add_collection(collection)
        ax.scatter(self.electrodes[:, 1], self.electrodes[:, 2])
        ax.autoscale_view()
        ax.set_aspect('equal')

        fig.savefig('test.png', dpi=300)
        return fig, ax

    def get_element_centroids(self):
        """return the central points of all elements

        Returns
        -------
        centroids: numpy.ndarray
            Nx2 array x/z coordinates for all (N) elements
        """
        if self.centroids is None:
            self.centroids = np.vstack((
                np.mean(self.grid['x'], axis=1),
                np.mean(self.grid['z'], axis=1)
            )).T

        return self.centroids

    @staticmethod
    def _get_area_polygon(points_x, points_z):
        """Return the area of a polygon. The node coordinates must be ordered,
        but the direction does not matter - we return the abs-value

        >>> points_x = [4,  4,  8,  8, -4, -4]
        >>> points_z = [6, -4, -4, -8, -8, 6]
        >>> self._get_area_polygon(points_x, points_z)
        ... 128
        """
        area = 0
        j = len(points_x) - 1
        for i in range(len(points_x)):
            area = area + (
                points_x[j] + points_x[i]
            ) * (points_z[j] - points_z[i])
            j = i
        return np.abs(area / 2)

    def get_element_areas(self):
        """return the areas of the elements

        note that this formula is generic, see CRMod code for a simpler one

        Returns
        -------
        areas : numpy.ndarray
        """
        areas = [
            self._get_area_polygon(
                points_x, points_z
            ) for points_x, points_z in zip(self.grid['x'], self.grid['z'])
        ]
        return np.array(areas)

    def get_electrode_positions(self):
        """Return the electrode positions in an numpy.ndarray

        Returns
        -------
        positions: numpy.ndarray
            Nx2 array, where N is the number of electrodes. The first column
            contains x positions, the second z positions.
        """
        return self.electrodes[:, 1:3]

    def get_min_max_electrode_distances(self):
        """Return the minimal and the maximal electrode distance of the grid

        Returns
        -------
        amin : float
            minimal electrode distance
        amax : float
            maximal electrode distance
        """
        distances = pdist(self.get_electrode_positions())
        return distances.min(), distances.max()

    def get_internal_angles(self):
        """Compute all internal angles of the grid

        Returns
        -------
        numpy.ndarray
            NxK array with N the number of elements, and K the number of nodes,
            filled with the internal angles in degrees
        """

        angles = []

        for elx, elz in zip(self.grid['x'], self.grid['z']):
            el_angles = []
            xy = np.vstack((elx, elz))
            for i in range(0, elx.size):
                i1 = (i - 1) % elx.size
                i2 = (i + 1) % elx.size

                a = (xy[:, i] - xy[:, i1])
                b = (xy[:, i2] - xy[:, i])
                # note that nodes are ordered counter-clockwise!
                angle = np.pi - np.arctan2(
                    a[0] * b[1] - a[1] * b[0],
                    a[0] * b[0] + a[1] * b[1]
                )
                el_angles.append(angle * 180 / np.pi)
            angles.append(el_angles)
        return np.array(angles)

    def analyze_internal_angles(self, return_plot=False):
        """Analyze the internal angles of the grid. Angles shouldn't be too
        small because this can cause problems/uncertainties in the
        Finite-Element solution of the forward problem. This function prints
        the min/max values, as well as quantiles, to the command line, and can
        also produce a histogram plot of the angles.

        Parameters
        ----------
        return_plot: bool
            if true, return (fig, ax) objects of the histogram plot

        Returns
        -------
        fig: matplotlib.figure
            figure object
        ax: matplotlib.axes
            axes object

        Examples
        --------

            >>> import crtomo.grid as CRGrid
                grid = CRGrid.crt_grid()
                grid.load_elem_file('elem.dat')
                fig, ax = grid.analyze_internal_angles(Angles)
            This grid was sorted using CutMcK. The nodes were resorted!
            Triangular grid found
            Minimal angle: 22.156368696965796 degrees
            Maximal angle: 134.99337326279496 degrees
            Angle percentile 10%: 51.22 degrees
            Angle percentile 20%: 55.59 degrees
            Angle percentile 30%: 58.26 degrees
            Angle percentile 40%: 59.49 degrees
            Angle percentile 50%: 59.95 degrees
            Angle percentile 60%: 60.25 degrees
            Angle percentile 70%: 61.16 degrees
            Angle percentile 80%: 63.44 degrees
            Angle percentile 90%: 68.72 degrees
            generating plot...
            >>> # save to file with
                fig.savefig('element_angles.png', dpi=300)

        """
        angles = self.get_internal_angles().flatten()

        print('Minimal angle: {0} degrees'.format(np.min(angles)))
        print('Maximal angle: {0} degrees'.format(np.max(angles)))
        # print out quantiles
        for i in range(10, 100, 10):
            print('Angle percentile {0}%: {1:0.2f} degrees'.format(
                i,
                np.percentile(angles, i),
            ))

        if return_plot:
            print('generating plot...')
            fig, ax = plt.subplots(1, 1, figsize=(12 / 2.54, 8 / 2.54))
            ax.hist(angles, int(angles.size / 10))
            ax.set_xlabel('angle [deg]')
            ax.set_ylabel('count')
            fig.tight_layout()
            # fig.savefig('plot_element_angles.jpg', dpi=300)
            return fig, ax

    @property
    def element_neighbors(self):
        """Return a list with element numbers (zero indexed) of neighboring
        elements. Note that the elements are not sorted. No spacial orientation
        can be inferred from the order of neighbors.

        WARNING: This function is slow due to a nested loop. This would be a
        good starting point for further optimizations.

        In order to speed things up, we could search using the raw data, i.e.,
        with CutMcK enabled sorting, and then restrict the loops to 2x the
        bandwidth (before - after).

        While not being returned, this function also sets the variable
        self.element_neighbors_edges, in which the common nodes with each
        neighbor are stored.

        Returns
        -------
        neighbors : list
            a list (length equal to nr of elements) with neighboring elements

        Examples
        --------


        """
        if self.element_neighbors_data is not None:
            return self.element_neighbors_data

        max_nr_edges = self.header['element_infos'][0, 2]

        # initialize the neighbor array
        self.element_neighbors_data = []
        self.element_neighbors_edges = []

        # determine neighbors
        print('Looking for neighbors')
        time_start = time.time()
        for nr, element_nodes in enumerate(self.elements):
            # print('element {0}/{1}'.format(nr + 1, self.nr_of_elements))
            # print(element_nodes)
            neighbors = []
            neighbors_edges = []  # store the edges to this neighbor
            for nr1, el in enumerate(self.elements):
                # we look for elements that have two nodes in common with this
                # element
                intersection = np.intersect1d(element_nodes, el)
                if intersection.size == 2:
                    neighbors.append(nr1)
                    neighbors_edges.append(intersection)
                    # stop if we reached the maximum number of possible edges
                    # this saves us quite some loop iterations
                    if len(neighbors) == max_nr_edges:
                        break
            self.element_neighbors_data.append(neighbors)
            self.element_neighbors_edges.append(neighbors_edges)
        time_end = time.time()
        print('elapsed time: {} s'.format(time_end - time_start))
        return self.element_neighbors_data

    def Wm(self):
        r"""Return the smoothing regularization matrix Wm of the grid

        See PhD Thesis Roland Blaschek, eq. 3.48ff

        ..math::

            \Psi_m = ||\underline{\underline{W}}_m \underline{m}||^2_2\\
            = \sum_{j=1}^M \sum_{i=nb(j)} \alpha_{r_ij} |\frac{m_j -
            m_i}{\Delta c_{ij}|^2 \Delta b_{ij} \Delta c_{ij}

        Note that the anisotropic regularization parameter :math:`\alpha` is
        not implemented yet.

        i and j are swapped in the implementation.

        See Fig 3.5 in the thesis.

        """
        centroids = self.get_element_centroids()

        Wm = scipy.sparse.lil_matrix(
            (self.nr_of_elements, self.nr_of_elements)
        )
        # Wm = np.zeros((self.nr_of_elements, self.nr_of_elements))
        for i, nb in enumerate(self.element_neighbors):
            for j, edges in zip(nb, self.element_neighbors_edges[i]):
                # side length
                edge_coords = self.nodes['presort'][edges][:, 1:]
                edge_length = np.linalg.norm(
                    edge_coords[1, :] - edge_coords[0, :]
                )
                distance = np.linalg.norm(centroids[i] - centroids[j])

                # main diagonal
                Wm[i, i] += np.sqrt(edge_length / distance)
                # side diagonals
                Wm[i, j] -= np.sqrt(edge_length / distance)
        return Wm

    def Wm_mgs(self, m, beta):
        r"""Return the MGS regularization matrix Wm of the grid

        See PhD Thesis Roland Blaschek, eq. 3.50

        Note that alpha and f are set 1

        Parameters
        ----------
        m : numpy.ndarray
            model parameters to compute MGS matrix for
        beta : float
            MGS beta parameter

        Returns
        -------
        Wm : scipy.sparse.csr_matrix
            Sparse MGS matrix
        """
        assert m is not None
        assert beta is not None
        centroids = self.get_element_centroids()

        Wm = scipy.sparse.csr_matrix(
            (self.nr_of_elements, self.nr_of_elements)
        )
        for j, nb in enumerate(self.element_neighbors):
            for i, edges in zip(nb, self.element_neighbors_edges[j]):
                # side length
                edge_coords = self.nodes['presort'][edges][:, 1:]
                # b_ij
                edge_length = np.linalg.norm(
                    edge_coords[1, :] - edge_coords[0, :]
                )
                # c_ij
                distance = np.linalg.norm(centroids[j] - centroids[i])

                # |model difference|
                m_diff = np.abs(m[i] - m[j])

                term = np.sqrt(
                    edge_length / distance * 1 / (
                        (m_diff / distance) ** 2 + beta ** 2
                    )
                )

                # main diagonal
                Wm[j, j] += term
                # side diagonals
                Wm[j, i] -= term
        return Wm

    @staticmethod
    def create_surface_grid(nr_electrodes=None,
                            spacing=None,
                            electrodes_x=None,
                            electrodes_z=None,
                            depth=None,
                            left=None,
                            right=None,
                            char_lengths=None,
                            lines=None,
                            internal_lines=None,
                            debug=False,
                            workdir=None,
                            force_neumann_only=False):
        """This is a simple wrapper for cr_trig_create to create simple surface
        grids.

        Electrode and boundary positions are rounded to the third digit.

        Parameters
        ----------
        nr_electrodes : int, optional
            the number of surface electrodes
        spacing : float, optional
            the spacing between electrodes, usually in [m], required if nr of
            electrodes is given
        electrodes_x : array, optional
            x-electrode positions can be provided here, e.g., for
            non-equidistant electrode distances
        electrodes_z : array, optional
            z-electrode positions can be provided here, e.g., for
            non-equidistant electrode distances. Only useful in combination
            with electrodes_x
        depth : float, optional
            the depth of the grid, relative to the minimum z-value.
            If not given, this is computed as half the
            maximum distance between electrodes
        left : float, optional
            the space allocated left of the first electrode. If not given,
            compute as a fourth of the maximum inter-electrode distance
        right : float, optional
            the space allocated right of the first electrode. If not given,
            compute as a fourth of the maximum inter-electrode distance
        char_lengths : float|list of 4 floats, optional
            characteristic lengths, as used by cr_trig_create
        lines: list of floats, optional
            at the given depths, add horizontal lines in the grid. Note that
            all positive values will be multiplied by -1!
        internal_lines : list of 4-tuple
            extra lines to add to the grid.  Important: These lines must NOT
            touch the outer edge of the grid (this is not supported by this
            function, but can be accomplished by manually building the grid)
        debug : bool, optional
            default: False. If true, don't hide the output of cr_trig_create
        workdir : string, optional
            if set, use this directory to create the grid. Don't delete files
            afterwards.
        force_neumann_only : bool, optional
            sometimes we want to use only Neumann boundary conditions. Setting
            this switch to True will force all boundaries to Neumann. Use only
            if you know what you are doing! default: False

        Returns
        -------
        grid: :class:`crtomo.grid.crt_grid` instance
            the generated grid

        Examples
        --------
        >>> from crtomo.grid import crt_grid
        >>> grid = crt_grid.create_surface_grid(40, spacing=0.25, depth=5,
        ...     left=2, right=2, char_lengths=[0.1, 0.5, 0.1, 0.5],
        ...     lines=[0.4, 0.8], debug=False, workdir=None)
        >>> import pylab as plt
        >>> fig, ax = plt.subplots()
        >>> grid.plot_grid_to_ax(ax)

        """
        # check if all required information are present
        if(electrodes_x is None and
           (nr_electrodes is None or spacing is None)):
            raise Exception(
                'You must provide either the parameter "electrodes_" or ' +
                'the parameters "nr_electrodes" AND "spacing"'
            )

        if electrodes_x is None:
            electrodes = np.array(
                [(x, 0.0) for x in np.arange(0.0, nr_electrodes)]
            )
            electrodes[:, 0] = electrodes[:, 0] * spacing
            electrodes = np.round(electrodes, 3)
        else:
            # we have individual electrode positions
            nr_electrodes = len(electrodes_x)
            electrodes = np.vstack(
                (electrodes_x, np.zeros_like(electrodes_x))
            ).T
            if electrodes_z is not None:
                electrodes[:, 1] = electrodes_z

        max_distance = np.abs(
            np.max(electrodes[:, 0]) - np.min(electrodes[:, 0])
        )
        minx = electrodes[:, 0].min()
        maxx = electrodes[:, 0].max()

        if left is None:
            left = max_distance / 4
        if right is None:
            right = max_distance / 4
        if depth is None:
            depth = np.abs(np.min(electrodes[:, 1]) - max_distance / 2)

        # min/max coordinates of final grid
        minimum_x = minx - left
        maximum_x = maxx + left
        minimum_z = np.min(electrodes[:, 1]) - depth
        # maximum_z = 0

        boundary_noflow = 11
        boundary_mixed = 12

        if force_neumann_only:
            boundary_mixed = 11

        # prepare extra lines
        extra_lines = []
        add_boundary_nodes_left = []
        add_boundary_nodes_right = []

        if electrodes_x is not None and lines is not None:
            raise Exception(
                'Extra lines not tested with individual electrode positions'
            )

        if lines is not None:
            lines = np.array(lines)
            lines[np.where(np.array(lines) < 0)] *= -1
            lines = sorted(lines)
            for line_depth in lines:
                extra_lines.append(
                    (minimum_x, -line_depth, maximum_x, -line_depth)
                )
                add_boundary_nodes_left.append(
                    (minimum_x, -line_depth, boundary_mixed)
                )
                add_boundary_nodes_right.append(
                    (maximum_x, -line_depth, boundary_mixed)
                )
            # reverse direction of nodes on the left side of the grid
            add_boundary_nodes_left = np.array(add_boundary_nodes_left)[::-1]
            add_boundary_nodes_right = np.array(add_boundary_nodes_right)
            print(add_boundary_nodes_left)
            print(add_boundary_nodes_right)

        if internal_lines is not None:
            extra_lines = extra_lines + internal_lines

        surface_electrodes = np.hstack((
            electrodes, boundary_noflow * np.ones((electrodes.shape[0], 1))
        ))

        # start assembling the boundaries here
        boundaries = np.vstack((
            # left corner
            (minimum_x, surface_electrodes[0, 1], boundary_noflow),
            # electrodes
            surface_electrodes,
            # right corner
            (maximum_x, surface_electrodes[-1, 1], boundary_mixed),
        ))

        if len(add_boundary_nodes_right) != 0:
            boundaries = np.vstack((
                boundaries,
                add_boundary_nodes_right,
            ))

        boundaries = np.vstack((
            boundaries,
            (maximum_x, minimum_z, boundary_mixed),
            (minimum_x, minimum_z, boundary_mixed),
        ))
        if len(add_boundary_nodes_left) != 0:
            boundaries = np.vstack(
                (
                    boundaries,
                    add_boundary_nodes_left,
                )
            )

        if char_lengths is None:
            assert spacing is not None, \
                "you must provide a spacing parameter or the char_lengths"
            char_lengths = [spacing / 3.0, ]
        else:
            assert isinstance(char_lengths, (float, int)) or len(
                    char_lengths) in [1, 4], \
                "char_lengths must be either of size 1 or size 4"

        if workdir is None:
            tempdir_obj = tempfile.TemporaryDirectory()
            tempdir = tempdir_obj.name
        else:
            if not os.path.isdir(workdir):
                os.makedirs(workdir)
            tempdir = workdir

        np.savetxt(
            tempdir + os.sep + 'electrodes.dat', electrodes,
            fmt='%.3f %.3f'
        )
        np.savetxt(tempdir + os.sep + 'boundaries.dat', boundaries,
                   fmt='%.3f %.3f %i')
        np.savetxt(
            tempdir + os.sep + 'char_length.dat',
            np.atleast_1d(char_lengths)
        )
        if extra_lines:
            np.savetxt(
                tempdir + os.sep + 'extra_lines.dat',
                np.atleast_2d(extra_lines),
                fmt='%.3f %.3f %.3f %.3f'
            )
        pwd = os.getcwd()
        os.chdir(tempdir)
        try:
            if debug:
                subprocess.call(
                    'cr_trig_create grid',
                    shell=True,
                )
            else:
                subprocess.check_output(
                    'cr_trig_create grid',
                    shell=True,
                    # stdout=subprocess.STDOUT,
                    # stderr=subprocess.STDOUT,
                )
        except subprocess.CalledProcessError as e:
            print('there was an error generating the grid')
            print(e.returncode)
            print(e.output)
            import shutil
            shutil.copytree(tempdir, pwd + os.sep + 'GRID_FAIL')
            exit()
        finally:
            os.chdir(pwd)
        grid = crt_grid(
            elem_file=tempdir + os.sep + 'grid' + os.sep + 'elem.dat',
            elec_file=tempdir + os.sep + 'grid' + os.sep + 'elec.dat',
        )
        if workdir is None:
            tempdir_obj.cleanup()

        return grid

    def get_element_indices_along_line(self, p0, p1, N):
        points = np.array(
            [(x, y) for x, y in zip(
                np.linspace(p0[0], p1[0], N), np.linspace(p0[1], p1[1], N)
            )]
        )
        centroids = self.get_element_centroids()

        tree = scipy.spatial.cKDTree(centroids)
        element_indices = tree.query(points)[1]
        return element_indices

    def get_element_indices_within_rectangle(self, xmin, xmax, zmin, zmax):
        """Return the indices of all elements whose center is located within
        the rectangle defined by the parameters.

        The indices can then be used, e.g., to select values from inversion
        results.

        Parameters
        ----------
        xmin : float
            Minimum x coordinate of accepted elements
        xmax : float
            Maximum x coordinate of accepted elements
        zmin : float
            Minimum z coordinate of accepted elements
        zmax : float
            Maximum z coordinate of accepted elements

        Returns
        -------
        indices : np.array
            Array with indices (zero-indexed)
        """
        centroids = self.get_element_centroids()
        indices_list = []
        for nr, (x, z) in enumerate(centroids):
            if x >= xmin and x <= xmax and z >= zmin and z <= zmax:
                indices_list.append(nr)
        return np.array(indices_list)

    @staticmethod
    def interpolate_grid(ingrid, outgrid, data, method='nearest'):
        """
        Function for interpolating data from one grid to another, using the
        cell midpoints as datapoint locations. Standard method for
        interpolating is set to nearest-neighbour.

        Parameters
        ----------
        ingrid : :class:`crtomo.grid.crt_grid` instance
            CRT grid that matches the input data
        outgrid : :class:`crtomo.grid.crt_grid` instance
            CRT grid to interpolate to
        data : pandas.dataframe
            input data that matches the input grid (in pandas dataframe format)
        method : interpolation method, optional
            Standard interpolation method is nearest-neighbour ('nearest').
            Other possible methods are 'linear' and 'cubic'.

        Returns
        -------
        interpolated data : pandas.dataframe
            returned dataframe with interpolated data
        """

        midpoints_in = ingrid.get_element_centroids()
        midpoints_out = outgrid.get_element_centroids()

        interpolated_data = pd.DataFrame()
        for i in list(data):
            interpolated_data[i] = scipy.interpolate.griddata(
                midpoints_in, data[i], midpoints_out, method=method)
        print(interpolated_data)
        return interpolated_data

    def reverse_node_order(self, element_type):
        """Reverses the order of the nodes of all elements of a given type.
        This can be used to fix CRTomo/CRMod errors regarding the computation
        of the determinant during FE system compilation.

        Use only if you know what you are doing!

        Parameters
        ----------
        element_type : int
            Element type to change. Usually 3 or 4 (triangles or quads)

        """
        assert element_type in self.element_data, \
            "element type {} not registered Available types :{}".format(
                element_type, self.element_data.keys()
            )
        for i in range(len(self.element_data[element_type])):
            self.element_data[
                element_type
            ][i].nodes = self.element_data[
                element_type
            ][i].nodes[::-1]

            self.element_data[
                element_type
            ][i].xcoords = self.element_data[
                element_type
            ][i].xcoords[::-1]

            self.element_data[
                element_type
            ][i].zcoords = self.element_data[
                element_type
            ][i].zcoords[::-1]



"""This is meant as a simple interface to CRMod, i.e. the forward operator and
its Jacobian. Nothing more.

It can be used to experiment with alternative inversion strategies outside of
the old FORTRAN code, but bear in mind that the CRMod binary is used to compute
everything. This is slow and involves writing things to disc. If possible use
another, better interfaced forward code such as pygimli.

The plan is to support both resistivity-only and complex (magnitude/phase)
versions.
"""
import os
import numpy as np

import crtomo
from crtomo.grid import crt_grid


class crmod_interface(object):
    def __init__(self, grid, configs, tempdir=None):
        """Initialize the CRMod interface

        Parameters
        ----------
        grid : crtomo.grid.crt_grid
            FE grid to work with
        configs : Nx4 numpy.ndarray
            Measurement configurations to work with (ABMN)
        tempdir : string, optional
            If set, use this directory for (temporary) file output. For example
            using the RAM-disc (/dev/shm) can potentially speed things up a
            lot.
        """
        assert isinstance(configs, np.ndarray)
        assert isinstance(grid, crt_grid)
        if tempdir is not None:
            assert os.path.isdir(tempdir)

        self.grid = grid
        self.configs = configs
        self.tempdir = tempdir

    def _get_tdm(self, m):
        """For a given model of resistivity magnitudes and phases, return a
        tdMan instance

        Parameters
        ----------
        m : Nx1|Nx2 ndarray
            N Model parameters, first column linear resistivity magnitudes [ohm
            m], second column resistivity phase values [mrad]

        Returns
        -------
        tdm : crtomo.tdMan
            td manager
        """
        if len(m.shape) == 1:
            m = m[:, np.newaxis]
        assert len(m.shape) == 2
        # print('gettm')
        # import IPython
        # IPython.embed()
        tdm = crtomo.tdMan(grid=self.grid, tempdir=self.tempdir)
        tdm.configs.add_to_configs(self.configs)

        pid_mag = tdm.parman.add_data(m[:, 0])
        tdm.register_magnitude_model(pid_mag)
        if m.shape[1] == 2:
            pid_pha = tdm.parman.add_data(m[:, 1])
        else:
            pid_pha = tdm.parman.add_data(np.zeros(m.shape[0]))
        tdm.register_phase_model(pid_pha)
        return tdm

    def forward_complex(self, log_sigma):
        r"""Compute a model response, i.e. complex impedances

        Parameters
        ----------
        log_sigma : 1xN or 2xN numpy.ndarray
            Model parameters. First column: :math:`log_e(\sigma)`, second
            column: :math:`\phi_{cond} [mrad]`.  If first dimension is of
            length one, assume phase values to be zero.

        Returns
        -------
        measurements : Nx2 numpy.ndarray
            Return log_e Y values of computed forward response
        """
        log_sigma = np.atleast_2d(log_sigma)
        rmag = 1.0 / np.exp(log_sigma[0, :])
        if log_sigma.shape[0] == 1:
            rpha = np.zeros_like(rmag)
        else:
            # convert to resistivities
            rpha = -log_sigma[1, :]
        m = np.vstack((rmag, rpha)).T
        tdm = self._get_tdm(m)
        measurements = tdm.measurements()
        # convert R to log Y
        measurements[:, 0] = np.log(1.0 / measurements[:, 0])
        # convert rpha to cpha
        measurements[:, 1] *= -1
        return measurements

    def J(self, log_sigma):
        """Return the sensitivity matrix

        Parameters
        ----------
        log_sigma : numpy.ndarray
            log_e conductivities

        """
        m = 1.0 / np.exp(log_sigma)
        tdm = self._get_tdm(m)

        tdm.model(
            sensitivities=True,
            # output_directory=stage_dir + 'modeling',
        )

        measurements = tdm.measurements()

        # build up the sensitivity matrix
        sens_list = []
        for config_nr, cids in sorted(
                tdm.assignments['sensitivities'].items()):
            sens_list.append(tdm.parman.parsets[cids[0]])

        # [del V / del sigma]
        sensitivities_lin = np.array(sens_list)

        # now convert to the log-sensitivities relevant for CRTomo and the
        # resolution matrix
        sensitivities_log = sensitivities_lin
        # multiply measurements on first dimension
        measurements_rep = np.repeat(
            measurements[:, 0, np.newaxis],
            sensitivities_lin.shape[1],
            axis=1)
        # sensitivities_log = sensitivities_log * mfit

        # multiply resistivities on second dimension
        m_rep = np.repeat(
            m[np.newaxis, :], sensitivities_lin.shape[0], axis=0
        )

        # eq. 3.41 in Kemna, 2000: notice that m_rep here is in rho, not sigma
        factor = - 1 / (m_rep * measurements_rep)
        sensitivities_log = factor * sensitivities_lin

        #         import IPython
        #         IPython.embed()

        return sensitivities_log

    def fwd_complex_R_sigma(self, model_ccomplex):
        """Compute the model response from linear complex conductivities

        Parameters
        ----------

        """
        m_rmag = 1.0 / np.abs(model_ccomplex)
        m_rpha = - np.arctan2(
            np.imag(model_ccomplex), np.real(model_ccomplex)
        ) * 1000
        model_rmag_rpha = np.vstack((m_rmag, m_rpha)).T
        tdm = self._get_tdm(model_rmag_rpha)
        measurements = tdm.measurements()
        rmag = measurements[:, 0]
        rpha = measurements[:, 1]

        rcomplex = rmag * np.exp(1j * rpha / 1000)
        return rcomplex

    def fwd_complex_R_rho(self, model_rcomplex):
        """Compute the forward response.

        Parameters
        ----------
        model_rcomplex : size M numpy.ndarray
            Complex forward model, complex resistivities

        Returns
        -------
        response : numpy.ndarray
            Model response, complex impedances

        """
        m_rmag = np.abs(model_rcomplex)
        m_rpha = np.arctan2(
            np.imag(model_rcomplex), np.real(model_rcomplex)
        ) * 1000
        model_rmag_rpha = np.vstack((m_rmag, m_rpha)).T
        tdm = self._get_tdm(model_rmag_rpha)
        measurements = tdm.measurements()
        rmag = measurements[:, 0]
        rpha = measurements[:, 1]

        rcomplex = rmag * np.exp(1j * rpha / 1000)
        return rcomplex

    def J_complex_R_sigma(self, model_ccomplex):
        """Compute the Jacobian

        """
        m_rmag = 1.0 / np.abs(model_ccomplex)
        m_rpha = - np.arctan2(
            np.imag(model_ccomplex), np.real(model_ccomplex)
        ) * 1000
        model_rmag_rpha = np.vstack((m_rmag, m_rpha)).T
        print('Reistivity model used to calculate Jacobian:')
        print(model_rmag_rpha)
        tdm = self._get_tdm(model_rmag_rpha)
        tdm.save_to_tomodir('tomodir')

        tdm.model(
            sensitivities=True,
            # output_directory=stage_dir + 'modeling',
        )

        # measurements = tdm.measurements()

        # build up the sensitivity matrix
        sens_list = []
        for config_nr, cids in sorted(
                tdm.assignments['sensitivities'].items()):
            print(cids)
            sens_re = tdm.parman.parsets[cids[0]]
            sens_im = tdm.parman.parsets[cids[1]]

            sens_complex = sens_re + 1j * sens_im
            # magnitude sensitvitiy
            sens_list.append(sens_complex)

        # [del V / del sigma]
        sensitivities_lin = np.array(sens_list)
        return sensitivities_lin

    def fwd_complex_logY_sigma(self, sigma):
        """Compute a model response from linear complex conductivities

        Parameters
        ----------
        sigma : 1xN or 2xN numpy.ndarray
            Model parameters as sigma and sigma phase, N the number of cells.
            If first dimension is of length one, assume phase values to be zero

        Returns
        -------
        measurements : Nx2 numpy nd array
            Return log_e sigma values of computed forward response (i.e., first
            column: log(sigma), second column: sigma phase
        """
        m_rmag = 1.0 / sigma
        tdm = self._get_tdm(m_rmag)
        measurements = tdm.measurements()
        # import IPython
        # IPython.embed()
        # convert R to log Y
        measurements[:, 0] = np.log(1.0 / measurements[:, 0])
        # convert resistivity phase to conductivity phase
        measurements[:, 1] *= -1
        return measurements

    def J_complex_Y_sigma(self, sigma_model_linear):
        r"""
        Compute the complex sensitivity matrix for :math:`\frac{\del V}{\del
        \sigma}`.

        Parameters
        ----------
        sigma_model_linear : numpy.ndarray
            complex linear conductivities

        Returns
        -------
        J_complex_lin : NxM numpy.ndarray
            Jacobian
        """

    def J_complex_logY_sigma(self, sigma):
        """Return the sensitivity matrix

        At this point works only with resistivities.

        Parameters
        ----------
        sigma : numpy.ndarray
            log_e conductivities

        """
        m_rmag = 1.0 / sigma
        tdm = self._get_tdm(m_rmag)

        tdm.model(
            sensitivities=True,
            output_directory='tmp_td'
            # output_directory=stage_dir + 'modeling',
        )

        measurements = tdm.measurements()

        # build up the sensitivity matrix
        sens_list = []
        for config_nr, cids in sorted(
                tdm.assignments['sensitivities'].items()):
            sens_list.append(tdm.parman.parsets[cids[0]])

        # [del V / del sigma]
        sensitivities_lin = np.array(sens_list)

        # multiply measurements on first dimension
        measurements_rep = np.repeat(
            measurements[:, 0, np.newaxis],
            sensitivities_lin.shape[1],
            axis=1)

        # after, eq. 3.41 in Kemna, 2000: notice that m_rep here is in rho, not
        # sigma
        # however, we now have model parameters in linear
        factor = -1 / measurements_rep
        sensitivities_logY = factor * sensitivities_lin

        return sensitivities_logY



#!/usr/bin/env python
"""Create depth-dependent grid decouplings.

WORK IN PROGRESS!
"""
from optparse import OptionParser

import numpy as np
from scipy.interpolate import UnivariateSpline

from crtomo.mpl_setup import *
import crtomo.grid as CRGrid


def handle_cmd_options():
    parser = OptionParser()
    parser.add_option('-e', "--elem", dest="elem_file", type="string",
                      help="elem.dat file (default: elem.dat)",
                      default="elem.dat")
    parser.add_option('-d', "--decfile", dest="dec_file", type="string",
                      help="elem.dat file (default: decfile.dat)",
                      default="decfile.dat")
    # parser.add_option("--eps", dest="eps", type="float",
    #                   help="User override for distance eps",
    #                   default=None)
    # parser.add_option('-l', "--linefile", dest="linefile",
    #                   help="Line file (default: extra_lines.dat)",
    #                   default='extra_lines.dat')
    parser.add_option("-o", "--output", dest="output",
                      help="Output file (default: decouplings.dat)",
                      metavar="FILE", default="decouplings.dat")

    (options, args) = parser.parse_args()
    return options


def load_decfile(decfile):
    decdata_raw = np.loadtxt(decfile)
    indices = np.argsort(decdata_raw[:, 0])
    decdata = decdata_raw[indices, :]

    return decdata


def get_univariate_spline(zn, dn):
    spl = UnivariateSpline(xn, dn)
    spl.set_smoothing_factor(0.1)
    return spl


def get_linear_fit(zn, dn):

    def func(z):
        if z <= zn.min():
            index = np.argmin(zn)
            return dn[index]
        if z >= zn.max():
            index = np.argmax(zn)
            return dn[index]
        if z in zn:
            index = np.where(z == zn)[0]
            return dn[index]

        diff = zn - z
        p = np.where(diff > 0)[0]
        pos_nr = np.argmin(zn[p])
        max_index = p[pos_nr]

        p = np.where(diff < 0)[0]
        neg_nr = np.argmax(zn[p])
        min_index = p[neg_nr]

        x = [zn[min_index], zn[max_index]]
        y = [dn[min_index], dn[max_index]]

        from scipy.interpolate import interp1d
        func = interp1d(x, y)

        return func(z)

    def array_func(z_array):
        result = [func(z) for z in z_array]
        return result

    return array_func


def get_func_decoupling_at_z(decfile, filename=None):
    decdata = load_decfile(decfile)

    # interpolate between the nearest z values
    # Use a spline interpolation to get a smooth curve
    zn = decdata[:, 0]
    dn = decdata[:, 1]

    # interp_func = get_univariate_spline(zn, dn)
    interp_func = get_linear_fit(zn, dn)

    # interp_func(-1.1)

    zs = np.linspace(zn.min(), zn.max(), 40)
    smooth_decouplings = interp_func(zs)

    if filename is not None:
        fig, ax = plt.subplots(1, 1, figsize=(5, 5))
        ax.plot(decdata[:, 1],
                decdata[:, 0],
                '.-',
                color='k')

        ax.plot(smooth_decouplings,
                zs,
                '.-',
                color='r')

        ax.set_xlabel('decoupling')
        ax.set_xlim(-0.1, 1.1)
        ax.set_ylabel('depth z')
        fig.tight_layout()
        fig.savefig(filename, dpi=300)
    return interp_func


def find_neighbors(grid, element_id):
    # find neighboring elements
    neighbors = []

    for index, elm1 in enumerate(grid.elements):
        indices = np.intersect1d(grid.elements[element_id], elm1)
        if len(indices) == 2:
            # append element index and the two adjoining nodes
            neighbors.append((
                index,
                indices[0],
                indices[1]
            ))
        # skip rest of the elements if we reached the maximum nr of possible
        # neighbors
        if len(neighbors) == 4:
            break
    return neighbors


def main():
    options = handle_cmd_options()
    decfunc = get_func_decoupling_at_z(
        options.dec_file,
        options.output + '.png'
    )
    grid = CRGrid.crt_grid()
    grid.load_elem_file(options.elem_file)

    nx = grid.nodes['sorted'][:, 1]
    ny = grid.nodes['sorted'][:, 2]
    nxy = np.vstack((nx, ny)).T

    decoupling_items = []
    for elmnr, element in enumerate(grid.elements):
        neighbors = find_neighbors(grid, elmnr)
        for neighbor in neighbors:
            xz = np.array([nxy[i] for i in neighbor[1:]])
            # we are only interested in z-coordinates
            z = xz[:, 1]

            if z.min() == z.max():
                # add decoupling
                decoupling = decfunc((z[0], ))[0]
                # print('decoupling', decoupling)
                decoupling_items.append((elmnr + 1,
                                         neighbor[0] + 1,
                                         decoupling))
            else:
                # print('vertical side')
                continue

    final_decouplings = np.array(decoupling_items)
    with open(options.output, 'w') as fid:
        fid.write('{0}\n'.format(final_decouplings.shape[0]))
        np.savetxt(fid, final_decouplings, fmt='%i %i %f')


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
"""
from optparse import OptionParser


def handle_cmd_options():
    parser = OptionParser()
    parser.add_option("-e", "--elem", dest="elem_file", type="string",
                      help="elem.dat file (default: elem.dat)",
                      default="elem.dat")
    parser.add_option("-t", "--elec", dest="elec_file", type="string",
                      help="elec.dat file (default: elec.dat)",
                      default="elec.dat")
    parser.add_option("--config", dest="config_file", type="string",
                      help="config.dat file (default: config.dat)",
                      default="config.dat")
    parser.add_option("-i", "--use_first_line", action="store_true",
                      dest="use_first_line", default=False,
                      help="Normally the first line of the config file is " +
                      "ignored, but if set to True, it will be used. " +
                      "Default: False")
    parser.add_option('-s', "--sink", dest="sink", type="int",
                      help="Fictitious sink node nr, implies 2D mode",
                      default=None)
    parser.add_option("--data", dest="data_file", type="string",
                      help="Data file (default: volt.dat)",
                      default='volt.dat')
    parser.add_option("-f", "--frequency", dest="frequency", type="int",
                      help="Frequency/Column in volt.dat, starting from 0 " +
                      "(default: 2)", default=2)

    parser.add_option("-o", "--output", dest="output_file", type="string",
                      help="Output file (plot) (default: sens_center.png)",
                      default='sens_center.png')

    parser.add_option("--cblabel", dest="cblabel", type="string",
                      help=r"ColorbarLabel (default: $Data$)",
                      default=r'$Data$')
    parser.add_option("--label", dest="label", type="string",
                      help=r"Label (default: none)", default=r'$ $')
    parser.add_option("-w", "--weight", dest="weight_int", type="int",
                      help="Choose the weights used : 0 - unweighted, 1 - " +
                      "abs, 2 -log10, 3 - sqrt", default=0)
    parser.add_option("-c", "--plot_configurations", action="store_true",
                      dest="plot_configurations", default=False,
                      help="Plots every configuration sensitivity center in " +
                      "a single file. Default: False")
    parser.add_option("--no_plot", action="store_true",
                      dest="no_plot", default=False,
                      help="Do not create center plot (only text output)")

    (options, args) = parser.parse_args()
    return options


def main():
    options = handle_cmd_options()
    print('NO PLOT', options.no_plot)



# *-* coding: utf-8 *-*
"""A digital representation of a sEIT inversion directory.

A sEIT-dir (or sipdir) has the following structure:
"""
from numbers import Number
import os
import shutil
from glob import glob

import pandas as pd
import numpy as np

import pylab as plt

from crtomo.grid import crt_grid
import crtomo.cfg as CRcfg
import crtomo.tdManager as CRman
# from crtomo.status import seitdir_is_finished

import sip_models.res.cc as cc_res
from reda.eis.plots import sip_response
# this is the same object as sip_response, but for legacy reasons we have
# multiple locations for it
from sip_models.sip_response import sip_response as sip_response2


class eitMan(object):
    """Manage sEIT data

    """
    def __init__(self, **kwargs):
        """
        Data/frequency input: Note that only one of the parameters seitdir,
        crt_data_dir, seit_data will be used during initialization, in the
        stated priority.

        Initialization diagram:

        .. blockdiag::

            diagram {
                group g_grid {
                    par_grid -> a_grid [label="loads"];
                    group g_elmc {
                        par_elem -> a_grid [label="loads"];
                        par_elec -> a_grid [label="loads"];
                    }
                }
                a_grid [label="grid"]
                seitdir[label="sEIT inversion directory"];
                seitdir -> a_grid [label="grid"];
                seitdir -> frequencies [label="loads"];
                seitdir -> inv_results [label="loads"];
                seitdir -> seit_data [label="loads"];
                crt_data_dir -> seit_data [label="loads"];
                inv_results[label="loads"];
                crt_data_dir[label="CRTomo style multi-frequency data"];
                crt_data_dir -> frequencies [label="loads"];
                crt_data_dir -> seit_data [label="loads"];
                dict_seit_data [label="data dictionary"];
                par_frequencies -> frequencies [label="loads"];

           }

        Parameters
        ----------
        seitdir : string, optional
            Load frequencies, grid, and data from an existing sEIT directory.
            Honors the shallow_import parameter.
        shallow_import : bool, optional
            In combination with seitdir, only import the last iteration
            inversion results (faster).
        crt_data_dir : string, optional
            if given, then try to load data from this directory. Expect a
            'frequencies.dat' file and corresponding 'volt_*.dat' files.
        only_frequency_ids : numpy.ndarray|None
            Load only the frequencies associated with these indices
            (zero-based). Only works with crt_data_dir.
        seit_data : dict
            A dictionary with frequencies as keys, and numpy arrays as items.
            Each array must have 6 columns: a,b,m,n,magnitude[Ohm],phase[mrad]
        frequencies : numpy.ndarray, optional
            frequencies that we work with
        grid : crtomo.crt_grid.crt_grid, optional
            A grid instance
        elem_file : string, optional
            Path to elem file
        elec_file : string, optional
            Path to elec file
        decouplings_file : str, optional
            Path to decoupling file: For not will be copied to the exe/ subdir
            of each tomodir written
        crmod_cfg : ?, optional

        crtomo_cfg : ?, optional

        parman : ?, optional

        """
        # the following variables will be partially duplicated in the tomodir
        # instances down below. Thus, they are used primarily to store blue
        # prints or "parent" files for the various frequencies. For example:
        # mask-files that are then parameterized using some kind of SIP model
        # such as the Cole-Cole model.
        self.crmod_cfg = kwargs.get('crmod_cfg', None)
        self.crtomo_cfg = kwargs.get('crtomo_cfg', CRcfg.crtomo_config())
        self.parman = kwargs.get('parman', None)
        self.noise_model = kwargs.get('noise_model', None)

        self.decoupling_file = kwargs.get('decouplings_file', None)

        # for each frequency we have a separate tomodir object
        self.tds = {}
        # when we load data from inversion results, we need to store the
        # corresponding parameter ids for the various quantities
        self.assigments = {
            'rmag': {},
            'rpha': {},
            'cre': {},
            'cim': {},
            'forward_rmag': {},
            'forward_rpha': {},
        }
        # shortcut
        self.a = self.assigments

        self.frequencies = kwargs.get('frequencies', None)

        # # now load data/initialize things
        seit_dir = kwargs.get('seitdir', None)
        crt_data_dir = kwargs.get('crt_data_dir', None)
        seit_data = kwargs.get('seit_data', None)

        # load/assign grid
        if 'grid' in kwargs:
            self.grid = kwargs.get('grid')
        elif 'elem_file' in kwargs and 'elec_file' in kwargs:
            grid = crt_grid()
            grid.load_grid(
                kwargs['elem_file'],
                kwargs['elec_file'],
            )
            self.grid = grid
        else:
            self.grid = None

        # these are the principle ways to add data
        if seit_dir is not None:
            # load the grid from the first invmod subdirectory
            tdirs = sorted(glob(seit_dir + '/invmod/*'))
            grid = crt_grid(
                elem_file=tdirs[0] + '/grid/elem.dat',
                elec_file=tdirs[0] + '/grid/elec.dat'
            )
            self.grid = grid

            self.load_inversion_results(
                seit_dir, shallow_import=kwargs.get('shallow_import', False)
            )
        elif crt_data_dir is not None:
            data_files = {}
            data_files['frequencies'] = '{}/frequencies.dat'.format(
                crt_data_dir)
            files = sorted(glob('{}/volt_*.crt'.format(crt_data_dir)))
            only_frequency_ids = kwargs.get('only_use_frequency_ids', None)
            data_files['crt'] = files
            self.load_data_crt_files(
                data_files, only_frequency_ids=only_frequency_ids
            )
        elif seit_data is not None:
            frequencies = sorted(seit_data.keys())
            self._init_frequencies(frequencies)

            for frequency in frequencies:
                abmn = seit_data[frequency][:, 0:4]
                rmag = seit_data[frequency][:, 4]
                rpha = seit_data[frequency][:, 5]

                self.tds[frequency].configs.add_to_configs(abmn)
                self.tds[frequency].register_measurements(rmag, rpha)
        else:
            # initialize frequency tomodirs
            if 'frequencies' in kwargs:
                self._init_frequencies(
                    kwargs.get('frequencies'),
                    kwargs.get('configs_abmn', None)
                )

        if self.grid is None:
            raise Exception(
                'You must provide either a grid instance or '
                'elem_file/elec_file file paths'
            )

    def _init_frequencies(self, frequencies, configs_abmn=None):
        """Initialize the tdMan instances associated with each frequency

        Note that existing tds will not be deleted/overwritten

        Parameters
        ----------
        frequencies : Nx1 numpy.ndarray
            Frequencies in ascending order
        configs_abmn : None|numpy.ndarray (Mx4)
            Measurement configurations to provide to the generated tdMan
            instances

        """
        self.frequencies = frequencies
        kwargs = {}

        for frequency in frequencies:
            if frequency in self.tds:
                # already present, do not generate a new tdMan instance
                continue
            if self.crtomo_cfg is not None:
                kwargs['crtomo_cfg'] = self.crtomo_cfg.copy()
            td = CRman.tdMan(
                grid=self.grid,
                configs_abmn=configs_abmn,
                **kwargs
            )
            self.tds[frequency] = td

    def set_area_to_sip_signature(self, xmin, xmax, zmin, zmax, spectrum):
        """Parameterize the eit instance by supplying one
        SIP spectrum and the area to apply to.

        Parameters
        ----------
        xmin : float
            Minimum x coordinate of the area
        xmax : float
            Maximum x coordinate of the area
        zmin : float
            Minimum z coordinate of the area
        zmax : float
            Maximum z coordinate of the area
        spectrum : sip_response
            SIP spectrum to use for parameterization

        """
        assert isinstance(spectrum, (sip_response, sip_response2))
        assert np.all(self.frequencies == spectrum.frequencies)
        for frequency, rmag, rpha in zip(
                self.frequencies, spectrum.rmag, spectrum.rpha):
            td = self.tds[frequency]
            pidm, pidp = td.a['forward_model']
            td.parman.modify_area(pidm, xmin, xmax, zmin, zmax, rmag)
            td.parman.modify_area(pidp, xmin, xmax, zmin, zmax, rpha)

    def set_area_to_single_colecole(self, xmin, xmax, zmin, zmax, ccpars):
        """Parameterize a rectangular area of the forward model using a
        single-term Cole-Cole response.

        Parameters
        ----------

        """
        objcc = cc_res.cc(self.frequencies)
        response = objcc.response(ccpars)
        self.set_area_to_sip_signature(xmin, xmax, zmin, zmax, response)

    def add_homogeneous_model(self, magnitude, phase=0, frequency=None):
        """Add homogeneous models to one or all tomodirs. Register those as
        forward models

        Parameters
        ----------
        magnitude : float
            Value of homogeneous magnitude model
        phase : float, optional
            Value of homogeneous phase model. Default 0
        frequency : float, optional
            Frequency of of the tomodir to use. If None, then apply to all
            tomodirs. Default is None.

        Returns
        -------
        None
        """
        if frequency is None:
            frequencies = self.frequencies
        else:
            assert isinstance(frequency, Number)
            frequencies = [frequency, ]

        for freq in frequencies:
            pidm, pidp = self.tds[freq].add_homogeneous_model(magnitude, phase)
            self.a['forward_rmag'][freq] = pidm
            self.a['forward_rpha'][freq] = pidp

    def register_forward_model(self, frequency, mag_model, phase_model):
        """Register a magnitude parameter set and optionally a phase parameter
        set as a forward model for a given frequency.

        Parameters
        ----------
        frequency : float
            Frequency. Must match the frequencies in eitMan.tds
        mag_model : numpy.ndarray
            The magnitude model (linear scale)
        phase_model : numpy.ndarray
            The phase model [mrad]
        """
        assert frequency in self.tds.keys(), "Frequency does not match any td"
        td = self.tds[frequency]

        pid_mag = td.register_magnitude_model(mag_model)
        pid_pha = td.register_phase_model(phase_model)
        self.a['forward_rmag'][frequency] = pid_mag
        self.a['forward_rpha'][frequency] = pid_pha

    def load_data_crt_files(self, data_dict, **kwargs):
        """Load sEIT data from .ctr files (volt.dat files readable by CRTomo,
        produced by CRMod)

        Parameters
        ----------
        data_dict : dict
            Data files that are imported. See example down below

        Examples
        --------

        >>> import glob
            data_files = {}
            data_files['frequencies'] = 'data/frequencies.dat'
            files = sorted(glob.glob('data/volt_*.crt'))
            data_files['crt'] = files

        """
        if isinstance(data_dict, str):
            raise Exception('Parameter must be a dict!')

        frequency_data = data_dict['frequencies']
        if isinstance(frequency_data, str):
            frequencies = np.loadtxt(data_dict['frequencies'])
        else:
            # if this is not a string, assume it to be the data
            frequencies = frequency_data

        if frequencies.size != len(data_dict['crt']):
            raise Exception(
                'number of frequencies does not match the number of data files'
            )
        only_frequency_ids = kwargs.get('only_frequency_ids', None)
        if only_frequency_ids is not None:
            frequencies = frequencies[only_frequency_ids]
            data_dict['crt'] = [
                data_dict['crt'][i] for i in only_frequency_ids
            ]
        self._init_frequencies(frequencies)

        for frequency, filename in zip(frequencies, data_dict['crt']):
            subdata = np.atleast_2d(np.loadtxt(filename, skiprows=1))
            if subdata.size == 0:
                continue
            # extract configurations
            A = (subdata[:, 0] / 1e4).astype(int)
            B = (subdata[:, 0] % 1e4).astype(int)
            M = (subdata[:, 1] / 1e4).astype(int)
            N = (subdata[:, 1] % 1e4).astype(int)

            ABMN = np.vstack((A, B, M, N)).T

            magnitudes = subdata[:, 2]
            phases = subdata[:, 3]

            self.tds[frequency].configs.add_to_configs(ABMN)
            self.tds[frequency].register_measurements(magnitudes, phases)

    def apply_crtomo_cfg(self):
        """Set the global crtomo_cfg for all frequencies
        """
        for key in sorted(self.tds.keys()):
            self.tds[key].crtomo_cfg = self.crtomo_cfg.copy()

    def apply_noise_models(self):
        """Set the global noise_model for all frequencies
        """
        for key in sorted(self.tds.keys()):
            self.tds[key].noise_model = self.noise_model

    def save_to_eitdir(self, directory):
        """Save the eit data into a eit/sip directory structure

        Parameters
        ----------
        directory: string|path
            output directory
        """
        if os.path.isdir(directory):
            raise Exception('output directory already exists')

        os.makedirs(directory)
        np.savetxt(directory + os.sep + 'frequencies.dat', self.frequencies)

        invmod_dir = directory + os.sep + 'invmod'
        os.makedirs(invmod_dir)
        for nr, key in enumerate(sorted(self.tds.keys())):
            outdir = invmod_dir + os.sep + '{0:02}_{1:.6f}'.format(nr, key)
            self.tds[key].save_to_tomodir(outdir)
            # HACK: if available, copy the decouplings file
            if self.decoupling_file is not None:
                shutil.copy(
                    self.decoupling_file,
                    outdir + os.sep + 'exe' + os.sep + 'decouplings.dat'
                )

    def reset_tds(self):
        """Reset the data stored in all tds"""
        for frequency, td in self.tds.items():
            td.reset_data()
        for key in ('rmag', 'rpha', 'cre', 'cim'):
            if key in self.a:
                self.a[key] = {}

    def load_inversion_results(self, sipdir, shallow_import=True):
        """Given an sEIT inversion directory, load inversion results and store
        the corresponding parameter ids in self.assignments

        Note that all previous data stored in this instance of the eitManager
        will be overwritten, if required!

        Parameters
        ----------
        sipdir : string
            path to a CRTomo sip-invdir (i.e., a full sEIT inversion directory)

        shallow_import : bool, optional
            If set to True, then only import the last inversion result, nothing
            else.
        """
        # load frequencies and initialize tomodir objects for all frequencies
        frequency_file = sipdir + os.sep + 'frequencies.dat'
        frequencies = np.loadtxt(frequency_file)
        self._init_frequencies(frequencies)

        # cycle through all tomodirs on disc and load the data
        for nr, (frequency_key, item) in enumerate(sorted(self.tds.items())):
            for label in ('rmag', 'rpha', 'cre', 'cim'):
                if label not in self.assigments:
                    self.a[label] = {}

            # some old inversion directories are one-indexed, while new ones
            # start with zero. As such, try both approaches
            for test_nr in (nr, nr + 1):
                tdir = ''.join((
                    sipdir + os.sep,
                    'invmod' + os.sep,
                    '{:02}_{:.6f}'.format(test_nr, frequency_key) + os.sep
                ))
                if os.path.isdir(tdir):
                    break
            if not os.path.isdir(tdir):
                raise Exception('tdir not found: {}'.format(tdir))

            if shallow_import:
                rmag_file = sorted(glob(tdir + 'inv/*.mag'))
                if len(rmag_file) > 0:
                    rmag_data = np.loadtxt(rmag_file[-1], skiprows=1)[:, 2]
                    pid_rmag = item.parman.add_data(rmag_data)
                else:
                    pid_rmag = None
                self.a['rmag'][frequency_key] = pid_rmag

                rpha_file = sorted(glob(tdir + 'inv/*.pha'))
                if len(rpha_file) > 0:
                    rpha_data = np.loadtxt(rpha_file[-1], skiprows=1)[:, 2]
                    pid_rpha = item.parman.add_data(rpha_data)
                else:
                    pid_rpha = None
                self.a['rpha'][frequency_key] = pid_rpha

                sigma_files = sorted(glob(tdir + 'inv/*.sig'))
                if len(sigma_files) > 0:
                    sigma_data = np.loadtxt(sigma_files[-1], skiprows=1)
                    pid_cre = item.parman.add_data(sigma_data[:, 0])
                    pid_cim = item.parman.add_data(sigma_data[:, 1])
                elif len(rmag_file) == 0 and len(rpha_file) == 0:
                    pid_cre = None
                    pid_cim = None
                else:
                    # very old CRTomo runs...
                    crho = item.parman.parsets[
                        pid_rmag
                    ] * np.exp(1j * item.parman.parsets[pid_rpha] / 1000)
                    csigma = 1 / crho
                    pid_cre = item.parman.add_data(csigma.real)
                    pid_cim = item.parman.add_data(csigma.imag)
                self.a['cre'][frequency_key] = pid_cre
                self.a['cim'][frequency_key] = pid_cim
            else:
                crtomo_cfg_file = tdir + os.sep + 'exe' + os.sep + 'crtomo.cfg'
                if os.path.isfile(crtomo_cfg_file):
                    item.crtomo_cfg.import_from_file(crtomo_cfg_file)
                # forward configurations
                config_file = tdir + os.sep + 'config' + os.sep + 'config.dat'
                if os.path.isfile(config_file):
                    item.configs.load_crmod_config(config_file)

                # load data/modeling results
                item._read_modeling_results(tdir + os.sep + 'mod')

                item.read_inversion_results(tdir)

                if len(item.a['inversion']['rmag']) > 0:
                    self.a['rmag'][frequency_key] = item.a[
                        'inversion']['rmag'][-1]
                    self.a['rpha'][frequency_key] = item.a[
                        'inversion']['rpha'][-1]
                    self.a['cre'][frequency_key] = item.a[
                        'inversion']['cre'][-1]
                    self.a['cim'][frequency_key] = item.a[
                        'inversion']['cim'][-1]
                else:
                    self.a['rmag'][frequency_key] = None
                    self.a['rpha'][frequency_key] = None
                    self.a['cre'][frequency_key] = None
                    self.a['cim'][frequency_key] = None

    def extract_polygon_area(self, label, polygon_points):
        """DEFUNCT

        Parameters
        ----------
        label : str
            the label (data type) to extract. This corresponds to a key in
            eitMan.assignments. Possible values are rmag, rpha, cre,
            cim
        polygon_points : list of (x,y) floats
            list of points that form the polygon

        Returns
        -------

        """
        raise Exception('Not implemented')
        """
        if isinstance(label, str):
            label = [label, ]

        data_list = []
        for label_key in label:
            value_list = {}
            for frequency, tdobj in sorted(self.tds.items()):
                values = tdobj.parman.extract_polygon_area(
                    pid=self.a[label_key][frequency],
                    polygon_points=polygon_points
                )
                value_list[key] = values
            df = pd.DataFrame(value_list)
            df['x'] = points[:, 0]
            df['z'] = points[:, 1]
            df['key'] = label_key
            data_list.append(df)
        df_all = pd.concat(data_list)

        return df_all
        """

    def extract_all_spectra(self, label):
        """Extract all SIP spectra, and return frequencies and a numpy array

        Parameters
        ----------
        label : str
            the label (data type) to extract. This corresponds to a key in
            eitMan.assignments. Possible values are rmag, rpha, cre, cim

        """
        if isinstance(label, str):
            label = [label, ]

        data_list = {}
        for label_key in label:
            frequencies = []
            value_list = []
            for frequency, item in sorted(self.tds.items()):
                pid = self.a[label_key][frequency]
                value_list.append(item.parman.parsets[pid])
                frequencies.append(frequency)
            data_list[label_key] = (frequencies, np.array(value_list))
        return data_list

    def extract_points(self, label, points):
        """Extract data points (i.e., SIP spectra) for one or more points.

        Parameters
        ----------
        label : str
            the label (data type) to extract. This corresponds to a key in
            eitMan.assignments. Possible values are rmag, rpha, cre,
            cim
        points : Nx2 numpy.ndarray
            (x, y) pairs

        Returns
        -------
        df_all : pandas.DataFrame
            A dataframe with the extracted data

        """
        if isinstance(label, str):
            label = [label, ]

        points = np.atleast_2d(points)

        data_list = []
        for label_key in label:
            value_list = {}
            for key, item in sorted(self.tds.items()):
                values = item.parman.extract_points(
                    pid=self.a[label_key][key],
                    points=points
                )
                value_list[key] = values
            df = pd.DataFrame(value_list)
            df['x'] = points[:, 0]
            df['z'] = points[:, 1]
            df['key'] = label_key
            data_list.append(df)
        df_all = pd.concat(data_list)

        return df_all

    def plot_forward_models(self, maglim=None, phalim=None, **kwargs):
        """Create plots of the forward models

        Returns
        -------
        return_dict : dict
            Dictionary containing the figure and axes objects of the magnitude
            plots

        """
        return_dict = {}

        N = len(self.frequencies)
        nrx = min(N, 3)
        nrz = int(np.ceil(N / nrx))

        labels = [
            r'$\rho~[\Omega m]$',
            r'$\phi~[mrad]$',
        ]

        cmaps = [
            'turbo',
            'jet',
        ]

        # try to select a suitable colorbar position
        (xmin, xmax), (zmin, zmax) = self.grid.get_minmax()
        if np.abs(xmax - xmin) > np.abs(zmax - zmin):
            cbposition = 'horizontal'
        else:
            cbposition = 'vertical'

        for index, key, limits in zip(
                (0, 1), ('rmag', 'rpha'), (maglim, phalim)):
            if limits is None:
                cbmin = None
                cbmax = None
            else:
                cbmin = limits[0]
                cbmax = limits[1]

            fig, axes = plt.subplots(
                nrz, nrx,
                figsize=(16 / 2.54, nrz * 3.5 / 2.54),
                sharex=True, sharey=True,
            )
            axes = np.atleast_2d(axes)
            for ax in axes.flat:
                ax.set_visible(False)

            for ax, frequency in zip(axes.flat, self.frequencies):
                ax.set_visible(True)
                td = self.tds[frequency]
                pids = td.a['forward_model']
                td.plot.plot_elements_to_ax(
                    pids[index],
                    ax=ax,
                    plot_colorbar=True,
                    cbposition=cbposition,
                    cbmin=cbmin,
                    cbmax=cbmax,
                    title='{:.3f} Hz'.format(frequency),
                    cblabel=labels[index],
                    cmap_name=cmaps[index],
                    **kwargs,
                )
            for ax in axes[0:-1, :].flat:
                ax.set_xlabel('')

            for ax in axes[:, 1:].flat:
                ax.set_ylabel('')

            fig.tight_layout()
            return_dict[key] = {
                'fig': fig,
                'axes': axes,
            }

        return return_dict

    def add_to_configs(self, configs):
        """Add configurations to all tomodirs

        Parameters
        ----------
        configs : :class:`numpy.ndarray`
            Nx4 numpy array with abmn configurations

        """
        for f, td in self.tds.items():
            td.configs.add_to_configs(configs)

    def model(self, **kwargs):
        """Run the forward modeling for all frequencies.

        Use :py:func:`crtomo.eitManager.eitMan.measurements` to retrieve the
        resulting synthetic measurement spectra.

        Parameters
        ----------
        **kwargs : dict, optional
            All kwargs are directly provide to the underlying
            :py:func:`crtomo.tdManager.tdMan.model` function calls.

        """
        for key, td in self.tds.items():
            td.model(**kwargs)

    def measurements(self, **kwargs):
        """Return modeled measurements. If not already done, call CRMod for
        each frequency to actually compute the forward response.

        Returns
        -------
        m_all : numpy.ndarray
            1. dimension: frequency
            2. dimension: config-number
            3. dimension: 2: magnitude and phase (resistivity)

        """
        m_all = np.array([self.tds[key].measurements(**kwargs) for key in
                          sorted(self.tds.keys())])
        return m_all

    def get_measurement_responses(self):
        """Return a dictionary of sip_responses for the modeled SIP spectra

        Note that this function does NOT check that each frequency contains the
        same configurations!

        Returns
        -------
        responses : dict
            Dictionary with configurations as keys

        """
        # take configurations from first tomodir
        configs = self.tds[sorted(self.tds.keys())[0]].configs.configs

        measurements = self.measurements()
        responses = {}
        for config, sip_measurement in zip(configs,
                                           np.rollaxis(measurements, 1)):
            sip = sip_response(
                frequencies=self.frequencies,
                rmag=sip_measurement[:, 0],
                rpha=sip_measurement[:, 1]
            )
            responses[tuple(config)] = sip
        return responses

    def save_measurements_to_directory(self, directory):
        """Store the measurements (either from a previous import, or from
        forward modeling, in one directory in the CRTomo 'crt'-style.

        Frequencies are stored in a file frequencies.dat.

        For each frequency, data is stored in a file volt_%.2i.dat in the
        CRTomo format.
        """
        os.makedirs(directory, exist_ok=True)
        np.savetxt(directory + os.sep + 'frequencies.dat', self.frequencies)
        for nr, (frequency, td) in enumerate(sorted(self.tds.items())):
            td.save_measurements(
                directory + os.sep + 'volt_{:02}_f{:08.3f}.dat'.format(
                    nr, frequency
                )
            )

    def plot_all_result_spectra(self, directory):
        os.makedirs(directory, exist_ok=True)
        N = self.tds[self.frequencies[0]].configs.configs.shape[0]
        for i in range(N):
            print('Plotting {}/{}'.format(i + 1, N))
            self.plot_result_spectrum(
                directory + '/spectrum_{:03}.jpg'.format(i), i)

    def plot_result_spectrum(self, filename, nr):
        """Plot a given data and inversion response spectrum to file.

        WARNING: At this point does not take into account missing
        configurations for certain frequencies...

        Parameters
        ----------
        filename : str
            Output filename
        nr : int
            Index of spectrum to plot. Starts at 0

        """
        rpha = np.vstack(
            [td.configs.measurements[
                td.a['inversion']['fwd_response_rpha'][-1]
            ] for f, td in sorted(self.tds.items())]).T
        rmag = np.vstack(
            [td.configs.measurements[
                td.a['inversion']['fwd_response_rmag'][-1]
            ] for f, td in sorted(self.tds.items())]).T
        spec1 = sip_response(
            self.frequencies, rmag=rmag[nr, :], rpha=rpha[nr, :])
        rmag_true = np.vstack(
            [td.configs.measurements[
                td.a['measurements'][0]
            ] for f, td in sorted(self.tds.items())]).T
        rpha_true = np.vstack(
            [td.configs.measurements[
                td.a['measurements'][1]
            ] for f, td in sorted(self.tds.items())]).T
        spec_true = sip_response(
            self.frequencies, rmag=rmag_true[nr, :], rpha=rpha_true[nr, :])
        spec_true.plot(
            filename,
            reciprocal=spec1,
            title='abmn: {}-{}-{}-{}'.format(
                *self.tds[self.frequencies[0]].configs.configs[nr]),
            label_nor='data',
            label_rec='inversion response',
        )

    def set_electrode_capacitances(self, capacitance):
        """Zimmermann et al 2018"""
        for frequency, td in self.tds.items():
            td.electrode_admittance = 2 * np.pi * frequency * capacitance

    def assign_sip_signatures_using_mask(self, mask_raw, lookup_table):
        """

        Parameters
        ----------
        lookup_table : dict

        spectrum : sip_response
            SIP spectrum to use for parameterization

        """
        assert isinstance(mask_raw, np.ndarray), "mask must be numpy array"
        # these are indices - we need them to be integers
        mask = mask_raw.astype(int)
        assert len(mask.shape) == 1, "mask must be an 1D array"
        assert (
            mask.size == self.grid.nr_of_elements
        ), "mask must be of the same size as number of mesh elements"

        assert isinstance(
            lookup_table, dict), "parameter lookup_table must be a dict"

        # check the lookup table
        for key, item in lookup_table.items():
            assert isinstance(
                item, (sip_response, sip_response2)
            ), "the item with key {} is not a sip_response!".format(key)
            assert np.all(self.frequencies == item.frequencies), \
                "The frequencies in the spectrum of key " + \
                "{} do not match".format(key)

        assert len(lookup_table) == np.unique(mask).size, \
            "Number of entries in the lookup table does not match number " + \
            "of unique mask entries"

        assert np.all(
            np.sort(
                np.unique(mask)
            ) == np.sort(np.unique(list(lookup_table.keys())))
            ), \
            "entries in mask and lookup_table do not match"

        if len(self.a['forward_rmag']) == 0:
            print('No forward models registered yet. Creating empty ones')
            self.add_homogeneous_model(0, 0)

        # loop over assignments
        for assignment in np.unique(mask):
            pixel_indices = np.where(mask == assignment)[0]

            spectrum = lookup_table[assignment]
            for frequency, rmag, rpha in zip(
                    self.frequencies, spectrum.rmag, spectrum.rpha):
                td = self.tds[frequency]
                pid_mag, pid_phase = td.a['forward_model']
                td.parman.modify_pixels(
                    pid_mag,
                    pixel_indices,
                    rmag
                )
                td.parman.modify_pixels(
                    pid_phase,
                    pixel_indices,
                    rpha
                )
